"""Test VeSync manager methods."""

import pytest
import asyncio
from unittest.mock import patch
import pyvesync
import logging
import copy
import time
import orjson
from itertools import chain
from pyvesync.vesyncfan import *
from pyvesync.devices.vesyncbulb import *
from pyvesync.devices.vesyncoutlet import *
from pyvesync.devices.vesyncpurifier import VeSyncAir131
from pyvesync.devices.vesyncswitch import *
import call_json as json_vals
from call_json_switches import SWITCHES_NUM
from call_json_outlets import OUTLETS_NUM
from call_json_fans import FANS_NUM
from call_json_bulbs import BULBS_NUM

BAD_DEV_LIST = {
    'result': {
        'total': 5,
        'pageNo': 1,
        'pageSize': 50,
        'list': [{'NoConfigKeys': None}],
    },
    'code': 0,
}


class TestDeviceList(object):
    """Test getting and populating device lists."""

    @pytest.fixture(scope='function')
    def api_mock(self, caplog):
        """Mock call_api and initialize VeSync object."""
        self.mock_api_call = patch('pyvesync.vesync.VeSync.async_call_api')
        self.mock_api = self.mock_api_call.start()
        self.loop = asyncio.new_event_loop()
        self.mock_api.return_value.ok = True
        self.vesync_obj = pyvesync.vesync.VeSync('sam@mail.com', 'pass', debug=True)
        self.vesync_obj.enabled = True
        self.vesync_obj.token = 'sample_tk'
        self.vesync_obj.account_id = 'sample_id'
        self.vesync_obj.in_process = False
        caplog.set_level(logging.DEBUG)
        yield
        self.mock_api_call.stop()

    async def run_coro(self, coro):
        """Run a coroutine in the event loop."""
        return await coro

    def run_in_loop(self, func, *args, **kwargs):
        """Run a function in the event loop."""
        return self.loop.run_until_complete(self.run_coro(func(*args, **kwargs)))

    def test_device_api(self, caplog, api_mock):
        """Tests to ensure call_api is being called correctly."""
        head = json_vals.DEFAULT_HEADER_BYPASS
        self.mock_api.return_value = (orjson.dumps({'V': 2}), 200)
        self.run_in_loop(self.vesync_obj.get_devices)
        call_list = self.mock_api.call_args_list
        call_p1 = call_list[0][0]
        call_p2 = call_list[0][1]
        assert call_p1[0] == '/cloud/v1/deviceManaged/devices'
        assert call_p1[1] == 'post'
        assert call_p2['headers'] == head
        assert self.vesync_obj.enabled

    @patch('pyvesync.vesyncoutlet.VeSyncOutlet7A')
    @patch('pyvesync.vesyncoutlet.VeSyncOutlet15A')
    @patch('pyvesync.vesyncoutlet.VeSyncOutlet10A')
    @patch('pyvesync.vesyncswitch.VeSyncWallSwitch')
    @patch('pyvesync.vesyncswitch.VeSyncDimmerSwitch')
    @patch('pyvesync.vesyncfan.VeSyncAir131')
    def test_getdevs_vsfact(
        self, air_patch, wsdim_patch, ws_patch, out10a_patch, out15a_patch, out7a_patch, api_mock
    ):
        """Test the get_devices, process_devices and VSFactory methods.
        Build list with device objects from details
        Test for all 6 known devices - 4 outlets, 2 switches, 1 fan.
        """

        device_list = json_vals.DeviceList.device_list_response()

        self.mock_api.return_value = device_list

        self.run_in_loop(self.vesync_obj.get_devices)

        assert len(self.vesync_obj.outlets) == OUTLETS_NUM
        assert len(self.vesync_obj.switches) == SWITCHES_NUM
        assert len(self.vesync_obj.fans) == FANS_NUM
        assert len(self.vesync_obj.bulbs) == BULBS_NUM

    def test_lv600(
            self, api_mock
    ):
        """Test the get_devices, process_devices and VSFactory methods.
        Build list with device objects from details
        Test for all 6 known devices - 4 outlets, 2 switches, 1 fan.
        """

        resp_dict, status = json_vals.DeviceList.FAN_TEST

        self.mock_api.return_value = orjson.dumps(resp_dict), status

        self.run_in_loop(self.vesync_obj.get_devices)

        assert len(self.vesync_obj.fans) == 3

    def test_dual200s(
            self, api_mock
    ):
        """Test the get_devices, process_devices and VSFactory methods.
        Build list with device objects from details
        Test for all 6 known devices - 4 outlets, 2 switches, 1 fan.
        """

        self.mock_api.return_value = json_vals.DeviceList.device_list_response('Dual200S')
        self.vesync_obj.debug = True
        self.run_in_loop(self.vesync_obj.get_devices)

        assert len(self.vesync_obj.fans) == 1

    def test_getdevs_code(self, caplog, api_mock):
        """Test get_devices with code > 0 returned."""
        device_list = (orjson.dumps({'code': 1, 'msg': 'gibberish'}), 200)

        self.mock_api.return_value = device_list
        self.run_in_loop(self.vesync_obj.get_devices)

        assert 'Error retrieving device list' in caplog.text

    def test_get_devices_resp_changes(self, caplog, api_mock):
        """Test if structure of device list response has changed."""
        resp_dict, status = (
            {
                'code': 0,
                'NOTresult': {
                    'NOTlist': [
                        {
                            'deviceType': 'wifi-switch-1.3',
                            'type': 'wifi-switch',
                            'cid': 'cid1',
                        }
                    ]
                },
            },
            200,
        )
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        self.run_in_loop(self.vesync_obj.get_devices)
        assert len(caplog.records) == 1
        assert 'Device list in response not found' in caplog.text

    def test_7a_bad_conf(self, caplog, api_mock):
        """Test bad device list response."""
        self.mock_api.return_value = (orjson.dumps(BAD_DEV_LIST), 200)
        self.run_in_loop(self.vesync_obj.get_devices)
        assert len(caplog.records) == 2

    def test_7a_no_dev_list(self, caplog, api_mock):
        """Test if empty device list is handled correctly."""
        empty_list = []

        self.vesync_obj.process_devices(empty_list)

        assert len(caplog.records) == 1

    def test_get_devices_devicetype_error(self, caplog, api_mock):
        """Test result and list keys exist but deviceType not in list."""
        resp_dict, status = (
            {'code': 0, 'result': {'list': [{'type': 'wifi-switch', 'cid': 'cid1'}]}},
            200,
        )
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        self.run_in_loop(self.vesync_obj.get_devices)
        assert len(caplog.records) == 2
        assert 'Error adding device' in caplog.text

    def test_unknown_device(self, caplog, api_mock):
        """Test unknown device type is handled correctly."""
        unknown_dev = json_vals.DeviceList.LIST_CONF_7A

        unknown_dev['devType'] = 'UNKNOWN-DEVTYPE'

        pyvesync.vesync.object_factory('unknown_device', unknown_dev, self.vesync_obj)

        assert len(caplog.records) == 1
        assert 'Unknown' in caplog.text

    def NO_test_time_check(self, api_mock):
        """Test device details update throttle."""
        # The throttle feature has been removed
        time_check = self.vesync_obj.device_time_check()
        assert time_check is True
        self.vesync_obj.last_update_ts = time.time()
        time_check = self.vesync_obj.device_time_check()
        assert time_check is False
        self.vesync_obj.last_update_ts = (
            time.time() - self.vesync_obj.update_interval - 1
        )
        time_check = self.vesync_obj.device_time_check()
        assert time_check is True

    @patch('pyvesync.vesyncoutlet.VeSyncOutlet7A', autospec=True)
    def test_remove_device(self, outlet_patch, caplog, api_mock):
        """Test remove device test."""
        device = copy.deepcopy(json_vals.DeviceList.LIST_CONF_7A)

        outlet_test = outlet_patch.return_value
        outlet_test.cid = '7A-CID'
        outlet_test.device_type = 'wifi-switch-1.3'
        outlet_test.device_name = '7A Device'

        new_list = [device]
        self.vesync_obj.outlets = [outlet_test]
        self.vesync_obj._remove_stale_devices(new_list)  # pylint: disable=protected-access

        assert self.vesync_obj.outlets == [outlet_test]

        device['cid'] = '7A-CID2'

        self.vesync_obj._remove_stale_devices([device])  # pylint: disable=protected-access

        assert not self.vesync_obj.outlets

        assert len(caplog.records) == 2

    @patch('pyvesync.vesyncoutlet.VeSyncOutdoorPlug', autospec=True)
    def test_add_dev_test(self, outdoor_patch, caplog, api_mock):
        """Test that new devices will not be added to instance."""
        outdoor_inst = VeSyncOutdoorPlug(
            json_vals.DeviceList.LIST_CONF_OUTDOOR_2, self.vesync_obj
        )
        self.vesync_obj.outlets = [outdoor_inst]

        add_test = self.vesync_obj._find_new_devices([json_vals.DeviceList.LIST_CONF_OUTDOOR_1])

        assert not add_test

    def test_display_func(self, caplog, api_mock):
        """Test display function outputs text."""
        self.vesync_obj.outlets.append(
            VeSyncOutdoorPlug(json_vals.DeviceList.device_list_item('ESO15-TB', 0), self.vesync_obj)
        )
        self.vesync_obj.outlets.append(
            VeSyncOutlet10A(json_vals.DeviceList.device_list_item('ESW01-EU'), self.vesync_obj)
        )
        self.vesync_obj.outlets.append(
            VeSyncOutlet15A(json_vals.DeviceList.device_list_item('ESW15-USA'), self.vesync_obj)
        )
        self.vesync_obj.outlets.append(
            VeSyncOutlet7A(json_vals.DeviceList.LIST_CONF_7A, self.vesync_obj)
        )
        self.vesync_obj.switches.append(
            VeSyncWallSwitch(json_vals.DeviceList.LIST_CONF_WS, self.vesync_obj)
        )
        self.vesync_obj.fans.append(
            VeSyncAir131(json_vals.DeviceList.LIST_CONF_AIR, self.vesync_obj)
        )
        self.vesync_obj.bulbs.append(
            VeSyncBulbESL100(json_vals.DeviceList.LIST_CONF_ESL100, self.vesync_obj)
        )

        dev_list = [
            self.vesync_obj.outlets,
            self.vesync_obj.switches,
            self.vesync_obj.fans,
            self.vesync_obj.bulbs,
        ]

        for device in chain(*dev_list):
            device.display()

        assert len(caplog.records) == 0

    @patch('pyvesync.vesyncoutlet.VeSyncOutlet7A', autospec=True)
    @patch('pyvesync.vesyncoutlet.VeSyncOutlet15A', autospec=True)
    @patch('pyvesync.vesyncoutlet.VeSyncOutlet10A', autospec=True)
    @patch('pyvesync.vesyncoutlet.VeSyncOutdoorPlug', autospec=True)
    @patch('pyvesync.vesyncbulb.VeSyncBulbESL100', autospec=True)
    @patch('pyvesync.vesyncswitch.VeSyncWallSwitch', autospec=True)
    @patch('pyvesync.vesyncfan.VeSyncAir131', autospec=True)
    def test_resolve_updates(
        self,
        air_patch,
        ws_patch,
        esl100_patch,
        outdoor_patch,
        out10a_patch,
        out15a_patch,
        out7a_patch,
        caplog,
        api_mock,
    ):
        """Test process_devices() with all devices.
        Creates vesync object with all devices and returns
        device list with new set of all devices.
        """
        out10a_patch.cid = '10A-CID1'
        out10a_patch.device_type = 'ESW10-EU'
        out10a_patch.device_name = '10A Removed'

        out15a_patch.cid = '15A-CID1'
        out15a_patch.device_type = 'ESW15-USA'
        out15a_patch.device_name = '15A Removed'

        out7a_patch.cid = '7A-CID1'
        out7a_patch.device_type = 'wifi-switch-1.3'
        out7a_patch.device_name = '7A Removed'

        outdoor_patch.cid = 'OUTDOOR-CID1'
        outdoor_patch.device_type = 'ESO15-TB'
        outdoor_patch.device_name = 'Outdoor Removed'

        esl100_patch.cid = 'BULB-CID1'
        esl100_patch.device_type = 'ESL100'
        esl100_patch.device_name = 'Bulb Removed'

        ws_patch.cid = 'WS-CID2'
        ws_patch.device_name = 'Switch Removed'
        ws_patch.device_type = 'ESWL01'

        air_patch.cid = 'AirCID2'
        air_patch.device_type = 'LV-PUR131S'
        air_patch.device_name = 'fan Removed'

        json_ret = json_vals.DeviceList.FULL_DEV_LIST

        self.vesync_obj.outlets.extend(
            [out7a_patch, out10a_patch, outdoor_patch]
        )
        self.vesync_obj.switches.extend([ws_patch])
        self.vesync_obj.fans.extend([air_patch])
        self.vesync_obj.bulbs.extend([esl100_patch])

        self.vesync_obj.process_devices(json_ret)

        assert len(self.vesync_obj.outlets) == 6
        assert len(self.vesync_obj.switches) == 2
        assert len(self.vesync_obj.fans) == 4
        assert len(self.vesync_obj.bulbs) == 4



"""
Air Purifier Device API Responses

FANS variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)
"""

from copy import deepcopy
from typing import Any
from pyvesync.device_map import purifier_modules
from pyvesync.const import ConnectionStatus, PurifierModes, DeviceStatus, AirQualityLevel
from defaults import (
    TestDefaults,
    FunctionResponses,
    build_bypass_v2_response,
    build_bypass_v1_response,
    FunctionResponsesV1,
    FunctionResponsesV2,
)


PURIFIER_MODELS = [m.setup_entry for m in purifier_modules]
# FANS = ['Core200S', 'Core300S', 'Core400S', 'Core600S', 'LV-PUR131S', 'LV600S',
#         'Classic300S', 'Classic200S', 'Dual200S', 'LV600S']
PURIFIERS_NUM = len(PURIFIER_MODELS)


class PurifierDefaults:
    purifier_mode = PurifierModes.MANUAL.value
    air_quality_enum = AirQualityLevel.GOOD
    fan_level = 1
    fan_set_level = 2
    filter_life = 80
    humidity = 50
    night_light = DeviceStatus.ON
    display = DeviceStatus.ON
    display_config = DeviceStatus.ON
    display_forever = True
    light_detection = False
    light_detected = True
    child_lock = DeviceStatus.ON
    air_quality_level = 1
    filter_open = 0
    aq_percent = 75
    air_quality_value_pm25 = 3
    pm1 = 10
    pm10 = 5
    rotate_angle = 45
    voc = 120
    co2 = 669
    temperature = 791


PURIFIER_DETAILS: dict[str, dict[str, Any]] = {
    "Core200S": {
        "enabled": True,
        "filter_life": PurifierDefaults.filter_life,
        "mode": PurifierDefaults.purifier_mode,
        "level": PurifierDefaults.fan_level,
        "display": bool(PurifierDefaults.display),
        "child_lock": bool(PurifierDefaults.child_lock),
        "night_light": PurifierDefaults.night_light.value,
        "configuration": {
            "display": bool(PurifierDefaults.display),
            "display_forever": PurifierDefaults.display_forever,
        },
        "device_error_code": 0,
    },
    "Core300S": {
        "enabled": True,
        "filter_life": PurifierDefaults.filter_life,
        "mode": PurifierDefaults.purifier_mode,
        "level": PurifierDefaults.fan_level,
        "air_quality": PurifierDefaults.air_quality_enum.value,
        "air_quality_value": PurifierDefaults.air_quality_value_pm25,
        "display": bool(PurifierDefaults.display),
        "child_lock": bool(PurifierDefaults.child_lock),
        "configuration": {
            "display": bool(PurifierDefaults.display_config),
            "display_forever": PurifierDefaults.display_forever,
            "auto_preference": {"type": "default", "room_size": 0},
        },
        "extension": {"schedule_count": 0, "timer_remain": 0},
        "device_error_code": 0,
    },
    "Core400S": {
        "enabled": True,
        "filter_life": PurifierDefaults.filter_life,
        "mode": PurifierDefaults.purifier_mode,
        "level": PurifierDefaults.fan_level,
        "air_quality": PurifierDefaults.air_quality_enum.value,
        "air_quality_value": PurifierDefaults.air_quality_value_pm25,
        "display": bool(PurifierDefaults.display),
        "child_lock": bool(PurifierDefaults.child_lock),
        "configuration": {
            "display": bool(PurifierDefaults.display_config),
            "display_forever": PurifierDefaults.display_forever,
            "auto_preference": {"type": "default", "room_size": 0},
        },
        "extension": {"schedule_count": 0, "timer_remain": 0},
        "device_error_code": 0,
    },
    "Core600S": {
        "enabled": True,
        "filter_life": PurifierDefaults.filter_life,
        "mode": PurifierDefaults.purifier_mode,
        "level": PurifierDefaults.fan_level,
        "air_quality": PurifierDefaults.air_quality_enum.value,
        "air_quality_value": PurifierDefaults.air_quality_value_pm25,
        "display": bool(PurifierDefaults.display),
        "child_lock": bool(PurifierDefaults.child_lock),
        "configuration": {
            "display": bool(PurifierDefaults.display_config),
            "display_forever": PurifierDefaults.display_forever,
            "auto_preference": {"type": "default", "room_size": 0},
        },
        "extension": {"schedule_count": 0, "timer_remain": 0},
        "device_error_code": 0,
    },
    "LV-PUR131S": {
        "screenStatus": PurifierDefaults.display.value,
        "filterLife": {
            "change": False,
            "useHour": 3520,
            "percent": PurifierDefaults.filter_life,
        },
        "activeTime": TestDefaults.active_time,
        "timer": None,
        "scheduleCount": 0,
        "schedule": None,
        "levelNew": PurifierDefaults.fan_set_level,
        "airQuality": str(PurifierDefaults.air_quality_enum),
        "level": PurifierDefaults.fan_level,
        "mode": PurifierDefaults.purifier_mode,
        "deviceName": "Levoit 131S Air Purifier",
        "currentFirmVersion": "2.0.58",
        "childLock": PurifierDefaults.child_lock.value,
        "deviceStatus": DeviceStatus.ON.value,
        "deviceImg": "https://image.vesync.com/defaultImages/deviceDefaultImages/airpurifier131_240.png",
        "connectionStatus": ConnectionStatus.ONLINE.value,
    },
    "LV-RH131S": {
        "enabled": bool(DeviceStatus.ON),
        "filter_life": PurifierDefaults.filter_life,
        "mode": PurifierDefaults.purifier_mode,
        "level": PurifierDefaults.fan_level,
        "air_quality": PurifierDefaults.air_quality_level,
        "display": PurifierDefaults.display,
        "child_lock": PurifierDefaults.child_lock,
        "buzzer": True,
        "configuration": {},
        "extension": {"schedule_count": 0, "timer_remain": 0},
    },
    "LAP-V102S": {  # Vital 100S
        "powerSwitch": int(DeviceStatus.ON),
        "filterLifePercent": PurifierDefaults.filter_life,
        "workMode": PurifierDefaults.purifier_mode,
        "manualSpeedLevel": PurifierDefaults.fan_set_level,
        "fanSpeedLevel": PurifierDefaults.fan_level,
        "AQLevel": int(PurifierDefaults.air_quality_enum),
        "PM25": PurifierDefaults.air_quality_value_pm25,
        "screenState": int(PurifierDefaults.display),
        "childLockSwitch": int(PurifierDefaults.child_lock),
        "screenSwitch": int(PurifierDefaults.display_config),
        "lightDetectionSwitch": int(PurifierDefaults.light_detection),
        "environmentLightState": int(PurifierDefaults.light_detected),
        "autoPreference": {"autoPreferenceType": "default", "roomSize": 600},
        "scheduleCount": 0,
        "timerRemain": 0,
        "efficientModeTimeRemain": 0,
        "errorCode": 0,
    },
    "LAP-V201S": {  # Vital 200S
        "powerSwitch": int(DeviceStatus.ON),
        "filterLifePercent": PurifierDefaults.filter_life,
        "workMode": PurifierDefaults.purifier_mode,
        "manualSpeedLevel": PurifierDefaults.fan_set_level,
        "fanSpeedLevel": PurifierDefaults.fan_level,
        "AQLevel": PurifierDefaults.air_quality_enum.value,
        "PM25": PurifierDefaults.air_quality_value_pm25,
        "screenState": PurifierDefaults.display,
        "childLockSwitch": PurifierDefaults.child_lock,
        "screenSwitch": PurifierDefaults.display_config,
        "lightDetectionSwitch": PurifierDefaults.light_detection,
        "environmentLightState": PurifierDefaults.light_detected,
        "autoPreference": {"autoPreferenceType": "default", "roomSize": 0},
        "scheduleCount": 0,
        "timerRemain": 0,
        "efficientModeTimeRemain": 0,
        "sleepPreference": {
            "sleepPreferenceType": "default",
            "cleaningBeforeBedSwitch": 1,
            "cleaningBeforeBedSpeedLevel": 3,
            "cleaningBeforeBedMinutes": 5,
            "whiteNoiseSleepAidSwitch": 1,
            "whiteNoiseSleepAidSpeedLevel": 1,
            "whiteNoiseSleepAidMinutes": 45,
            "duringSleepSpeedLevel": 5,
            "duringSleepMinutes": 480,
            "afterWakeUpPowerSwitch": 1,
            "afterWakeUpWorkMode": "auto",
            "afterWakeUpFanSpeedLevel": 1,
        },
        "errorCode": 0,
    },
    "EL551S": {  # Everest Air
        "fanRotateAngle": PurifierDefaults.rotate_angle,
        "filterOpenState": PurifierDefaults.filter_open,
        "powerSwitch": int(DeviceStatus.ON),
        "filterLifePercent": PurifierDefaults.filter_life,
        "workMode": PurifierDefaults.purifier_mode,
        "manualSpeedLevel": PurifierDefaults.fan_set_level,
        "fanSpeedLevel": PurifierDefaults.fan_level,
        "AQLevel": PurifierDefaults.air_quality_enum.value,
        "AQPercent": PurifierDefaults.aq_percent,
        "PM25": PurifierDefaults.air_quality_value_pm25,
        "PM1": PurifierDefaults.pm1,
        "PM10": PurifierDefaults.pm10,
        "screenState": int(PurifierDefaults.display),
        "childLockSwitch": int(PurifierDefaults.child_lock),
        "screenSwitch": int(PurifierDefaults.display_config),
        "lightDetectionSwitch": int(PurifierDefaults.light_detection),
        "environmentLightState": int(PurifierDefaults.light_detected),
        "autoPreference": {"autoPreferenceType": "default", "roomSize": 0},
        "routine": {"routineType": "normal", "runSeconds": 0},
        "scheduleCount": 0,
        "timerRemain": 0,
        "efficientModeTimeRemain": 0,
        "ecoModeRunTime": 0,
        "errorCode": 0,
    },
    "LAP-B851S-WUS": {
        "powerSwitch": TestDefaults.bin_toggle,
        "workMode": PurifierDefaults.purifier_mode,
        "manualSpeedLevel": PurifierDefaults.fan_set_level,
        "fanSpeedLevel": PurifierDefaults.fan_level,
        "PM25": PurifierDefaults.air_quality_value_pm25,
        "PM1": PurifierDefaults.pm1,
        "PM10": PurifierDefaults.pm10,
        "screenState": int(PurifierDefaults.display),
        "childLockSwitch": int(PurifierDefaults.child_lock),
        "screenSwitch": int(PurifierDefaults.display_config),
        "lampType": 0,
        "roomSize": 242,
        "lampSwitch": int(PurifierDefaults.display),
        "autoPreference": {"autoPreferenceType": "default", "roomSize": 630},
        "scheduleCount": 0,
        "timerRemain": 0,
        "efficientModeTimeRemain": 1182,
        "humidity": PurifierDefaults.humidity,
        "AQI": PurifierDefaults.aq_percent,
        "AQLevel": PurifierDefaults.air_quality_level,
        "VOC": PurifierDefaults.voc,
        "CO2": PurifierDefaults.co2,
        "temperature": PurifierDefaults.temperature,
        "nightLight": {
            "nightLightSwitch": int(PurifierDefaults.night_light),
            "brightness": TestDefaults.brightness,
            "colorTemperature": TestDefaults.color_temp_k,
        },
        "breathingLamp": {
            "breathingLampSwitch": TestDefaults.bin_toggle,
            "colorTemperature": TestDefaults.color_temp_k,
            "timeInterval": 5,
            "brightnessStart": 10,
            "brightnessEnd": 90,
        },
        "errorCode": 0,
        "dumpedState": 0,
        "whiteNoiseInfo": {
            "playStatus": 0,
            "soundId": 100006,
            "countDown": 1800,
            "countingDown": 1800,
            "downloadStatus": 2,
        },
        "guardingInfo": {"guarding": 0, "remainTS": 100},
    },
}


DETAILS_RESPONSES = {
    "LV-PUR131S": build_bypass_v1_response(result_dict=PURIFIER_DETAILS["LV-PUR131S"]),
    "LV-RH131S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["LV-RH131S"]),
    "Core200S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["Core200S"]),
    "Core300S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["Core300S"]),
    "Core400S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["Core400S"]),
    "Core600S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["Core600S"]),
    "LAP-V102S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["LAP-V102S"]),
    "LAP-V201S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["LAP-V201S"]),
    "EL551S": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["EL551S"]),
    "LAP-B851S-WUS": build_bypass_v2_response(inner_result=PURIFIER_DETAILS["LAP-B851S-WUS"]),
}

FunctionResponses.default_factory = lambda: (
    {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {"traceId": TestDefaults.trace_id, "code": 0},
    },
    200,
)

METHOD_RESPONSES = {
    "LV-PUR131S": deepcopy(FunctionResponsesV1),
    "LV-RH131S": deepcopy(FunctionResponsesV2),
    "Core200S": deepcopy(FunctionResponsesV2),
    "Core300S": deepcopy(FunctionResponsesV2),
    "Core400S": deepcopy(FunctionResponsesV2),
    "Core600S": deepcopy(FunctionResponsesV2),
    "LAP-V102S": deepcopy(FunctionResponsesV2),
    "LAP-V201S": deepcopy(FunctionResponsesV2),
    "EL551S": deepcopy(FunctionResponsesV2),
    "LAP-B851S-WUS": deepcopy(FunctionResponsesV2),
}

# Add responses for methods with different responses than the default

# # Timer Responses

# for k in AIR_MODELS:
#     METHOD_RESPONSES[k]["set_timer"] = (INNER_RESULT({"id": 1}), 200)
#     METHOD_RESPONSES[k]["get_timer"] = (
#         INNER_RESULT({"id": 1, "remain": 100, "total": 100, "action": "off"}),
#         200,
#     )
# FAN_TIMER = helpers.Timer(100, "off")



"""Etekcity Outlets."""

from __future__ import annotations

import calendar
import logging
from datetime import UTC, datetime
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.outlet_base import VeSyncOutlet
from pyvesync.const import (
    ENERGY_HISTORY_OFFSET_WHOGPLUG,
    STATUS_OK,
    ConnectionStatus,
    DeviceStatus,
    EnergyIntervals,
)
from pyvesync.models.base_models import RequestHeaders
from pyvesync.models.bypass_models import (
    BypassV2InnerResult,
    RequestBypassV1,
    TimerModels,
)
from pyvesync.models.outlet_models import (
    Request15ADetails,
    Request15ANightlight,
    Request15AStatus,
    RequestESW03Status,
    RequestOutdoorStatus,
    RequestWHOGYearlyEnergy,
    Response7AOutlet,
    Response15ADetails,
    ResponseBSDGO1OutletResult,
    ResponseEnergyResult,
    ResponseESW03Details,
    ResponseOutdoorDetails,
    ResponseWHOGResult,
    ResultESW10Details,
    Timer7AItem,
)
from pyvesync.utils.device_mixins import (
    BypassV1Mixin,
    BypassV2Mixin,
    process_bypassv1_result,
    process_bypassv2_result,
)
from pyvesync.utils.helpers import Helpers, Timer
from pyvesync.utils.logs import LibraryLogger

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import OutletMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

logger = logging.getLogger(__name__)


class VeSyncOutlet7A(VeSyncOutlet):
    """Etekcity 7A Round Outlet Class.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize Etekcity 7A round outlet class."""
        self.request_keys: list[str] = []
        super().__init__(details, manager, feature_map)

    def _build_headers(self) -> dict:
        """Build 7A Outlet Request Headers."""
        headers = RequestHeaders.copy()
        headers.update(
            {
                'tz': self.manager.time_zone,
                'tk': self.manager.token,
                'accountid': self.manager.account_id,
            }
        )
        return headers

    async def get_details(self) -> None:
        r_dict, _ = await self.manager.async_call_api(
            '/v1/device/' + self.cid + '/detail',
            'get',
            headers=self._build_headers(),
        )

        if not isinstance(r_dict, dict):
            LibraryLogger.error_device_response_content(
                logger,
                self,
                'get_details',
                'Response is not valid JSON',
            )
            return

        if 'error' in r_dict:
            _ = Helpers.process_dev_response(logger, 'get_details', self, r_dict)
            return
        self.state.update_ts()
        resp_model = Response7AOutlet.from_dict(r_dict)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.device_status = resp_model.deviceStatus
        self.state.active_time = resp_model.activeTime
        self.state.energy = resp_model.energy
        self.state.power = self.parse_energy_detail(resp_model.power)
        self.state.voltage = self.parse_energy_detail(resp_model.voltage)

    @staticmethod
    def parse_energy_detail(energy: str | float) -> float:
        """Parse energy details to be compatible with new and old firmware."""
        try:
            if isinstance(energy, str) and ':' in energy:
                power = round(float(Helpers.calculate_hex(energy)), 2)
            else:
                power = float(energy)
        except ValueError:
            logger.warning('Error parsing power response - %s', energy)
            power = 0
        return power

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        toggle_str = DeviceStatus.ON if toggle else DeviceStatus.OFF
        r_dict, status_code = await self.manager.async_call_api(
            f'/v1/wifi-switch-1.3/{self.cid}/status/{toggle_str}',
            'put',
            headers=Helpers.req_legacy_headers(self.manager),
        )

        if status_code != STATUS_OK:
            LibraryLogger.error_device_response_content(
                logger,
                self,
                'toggle_switch',
                'Response code is not 200',
            )
            return False

        if isinstance(r_dict, dict) and 'error' in r_dict:
            _ = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
            return False

        self.state.update_ts()
        self.state.device_status = toggle_str
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        r_dict, status_code = await self.manager.async_call_api(
            f'/v2/device/{self.cid}/timer',
            'get',
            headers=Helpers.req_legacy_headers(self.manager),
        )
        if not r_dict or status_code != STATUS_OK:
            logger.debug('No timer set.')
            self.state.timer = None
            return
        if isinstance(r_dict, list) and len(r_dict) > 0:
            timer_model = Helpers.model_maker(logger, Timer7AItem, 'get_timer', r_dict[0])
            if timer_model is None:
                self.state.timer = None
                return
            self.state.timer = Timer(
                timer_duration=int(timer_model.counterTimer),
                id=int(timer_model.timerID),
                action=timer_model.action,
            )
            if timer_model.timerStatus == 'off':
                self.state.timer.pause()
            return
        self.state.timer = None

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if not isinstance(action, str) or action not in [
            DeviceStatus.ON,
            DeviceStatus.OFF,
        ]:
            logger.error('Invalid action for timer - %s', action)
            return False
        update_dict = {
            'action': action,
            'counterTimer': duration,
            'timerStatus': 'start',
            'conflictAwayIds': [],
            'conflictScheduleIds': [],
            'conflictTimerIds': [],
        }
        r_dict, status_code = await self.manager.async_call_api(
            f'/v2/device/{self.cid}/timer',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=update_dict,
        )
        if status_code != STATUS_OK or not isinstance(r_dict, dict):
            logger.warning('Failed to set timer for %s.', self.device_name)
            return False

        if 'error' in r_dict:
            logger.warning('Error in response: %s', r_dict['error'])
            return False

        result_model = Helpers.model_maker(
            logger, TimerModels.ResultV1SetTimer, 'set_timer', r_dict, self
        )
        if result_model is None:
            logger.warning('Failed to set timer.')
            return False
        if result_model.timerID == '':
            logger.warning('Unable to set timer.')
            if result_model.conflictTimerIds:
                logger.warning(
                    'Conflicting timer IDs - %s', result_model.conflictTimerIds
                )
            return False
        self.state.timer = Timer(duration, action, int(result_model.timerID))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set, nothing to clear, run get_timer().')
            return False
        _, status_code = await self.manager.async_call_api(
            f'/v2/device/{self.cid}/timer/{self.state.timer.id}',
            'delete',
            headers=Helpers.req_legacy_headers(self.manager),
        )
        if status_code != STATUS_OK:
            return False
        self.state.timer = None
        return True


class VeSyncOutlet10A(BypassV1Mixin, VeSyncOutlet):
    """Etekcity 10A ESW01 and ESW03 Round Outlets.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize 10A outlet class."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            RequestBypassV1, method='deviceDetail', endpoint='deviceDetail'
        )
        r = process_bypassv1_result(
            self, logger, 'get_details', r_dict, ResponseESW03Details
        )
        if r is None:
            return

        self.state.device_status = DeviceStatus(r.deviceStatus)
        self.state.connection_status = ConnectionStatus(r.connectionStatus)
        self.state.energy = r.energy or 0
        self.state.power = r.power or 0
        self.state.voltage = r.voltage or 0

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        toggle_str = DeviceStatus.ON if toggle else DeviceStatus.OFF
        update_dict = {'status': toggle_str.value}
        response = await self.call_bypassv1_api(
            RequestESW03Status,
            update_dict=update_dict,
            method='deviceStatus',
            endpoint='deviceStatus',
        )
        r_dict = Helpers.process_dev_response(logger, 'toggle_switch', self, response)
        if r_dict is None:
            return False

        self.state.device_status = toggle_str
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncOutlet15A(BypassV1Mixin, VeSyncOutlet):
    """Class for Etekcity 15A Rectangular Outlets.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize 15A rectangular outlets."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            Request15ADetails, method='deviceDetail', endpoint='deviceDetail'
        )

        r = Helpers.process_dev_response(logger, 'get_details', self, r_dict)
        if r is None:
            return

        resp_model = Response15ADetails.from_dict(r)
        result = resp_model.result
        self.state.device_status = result.deviceStatus
        self.state.connection_status = result.connectionStatus
        self.state.nightlight_status = result.nightLightStatus
        self.state.nightlight_brightness = result.nightLightBrightness
        self.state.nightlight_automode = result.nightLightAutoMode
        self.state.active_time = result.activeTime
        self.state.power = result.power or 0
        self.state.voltage = result.voltage or 0
        self.state.energy = result.energy or 0

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        toggle_str = DeviceStatus.ON if toggle else DeviceStatus.OFF
        r_dict = await self.call_bypassv1_api(
            Request15AStatus,
            update_dict={'status': toggle_str},
            method='deviceStatus',
            endpoint='deviceStatus',
        )
        response = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if response is None:
            return False

        self.state.device_status = toggle_str
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def set_nightlight_state(self, mode: str) -> bool:
        """Set nightlight state for 15A Outlets."""
        if mode.lower() not in self.nightlight_modes:
            logger.error('Invalid nightlight mode - %s', mode)
            return False
        mode = mode.lower()
        r_dict = await self.call_bypassv1_api(
            Request15ANightlight,
            update_dict={'mode': mode},
            method='outletNightLightCtl',
            endpoint='outletNightLightCtl',
        )

        response = Helpers.process_dev_response(
            logger, 'set_nightlight_state', self, r_dict
        )
        if response is None:
            return False

        self.state.nightlight_status = mode
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        method = 'getTimers'
        endpoint = f'/timer/{method}'
        r_dict = await self.call_bypassv1_api(
            Request15ADetails, method=method, endpoint=endpoint
        )
        result_model = process_bypassv1_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return
        timers = result_model.timers
        if not isinstance(timers, list) or len(timers) == 0:
            self.state.timer = None
            return
        timer = timers[0]
        if not isinstance(timer, TimerModels.TimerItemV1):
            logger.warning('Invalid timer model - %s', timer)
            return
        self.state.timer = Timer(
            timer_duration=int(timer.counterTimer),
            id=int(timer.timerID),
            action=timer.action,
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if not action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            logger.error('Invalid action for timer - %s', action)
            return False
        update_dict = {
            'action': action,
            'counterTime': str(duration),
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime,
            update_dict=update_dict,
            method='addTimer',
            endpoint='timer/addTimer',
        )
        result_model = process_bypassv1_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(duration, action, int(result_model.timerID))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set, nothing to clear, run get_timer().')
            return False
        if self.state.timer.time_remaining == 0:
            logger.debug('Timer already ended.')
            self.state.timer = None
            return True
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            {'timerId': str(self.state.timer.id)},
            method='deleteTimer',
            endpoint='timer/deleteTimer',
        )
        r_dict = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r_dict is None:
            if (
                self.last_response is not None
                and self.last_response.name == 'TIMER_NOT_EXISTS'
            ):
                self.state.timer = None
            return False
        self.state.timer = None
        return True


class VeSyncOutdoorPlug(BypassV1Mixin, VeSyncOutlet):
    """Class to hold Etekcity outdoor outlets.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize Etekcity Outdoor Plug class."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv1_api(
            Request15ADetails, method='deviceDetail', endpoint='deviceDetail'
        )
        r = Helpers.process_dev_response(logger, 'get_details', self, r_dict)
        if r is None:
            return

        resp_model = ResponseOutdoorDetails.from_dict(r)
        self.state.connection_status = resp_model.result.connectionStatus
        self.state.energy = resp_model.result.energy
        self.state.power = resp_model.result.power
        self.state.voltage = resp_model.result.voltage
        self.state.active_time = resp_model.result.activeTime
        for outlet in resp_model.result.subDevices:
            if not isinstance(self.sub_device_no, float):
                continue
            if int(self.sub_device_no) == int(outlet.subDeviceNo):
                self.state.device_status = outlet.subDeviceStatus

    @deprecated('Use toggle_switch(toggle: bool | None) instead')
    async def toggle(self, status: str) -> bool:
        """Deprecated - use toggle_switch()."""
        toggle = status != DeviceStatus.ON
        return await self.toggle_switch(toggle)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        status = DeviceStatus.ON if toggle else DeviceStatus.OFF

        r_dict = await self.call_bypassv1_api(
            RequestOutdoorStatus,
            update_dict={'switchNo': self.sub_device_no, 'status': status},
            method='deviceStatus',
            endpoint='deviceStatus',
        )

        response = Helpers.process_dev_response(logger, 'toggle', self, r_dict)
        if response is None:
            return False

        self.state.device_status = status
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> None:
        method = 'getTimers'
        endpoint = f'/timer/{method}'
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1GetTimer,
            {'switchNo': self.sub_device_no},
            method=method,
            endpoint=endpoint,
        )
        result_model = process_bypassv1_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV1GetTimer
        )
        if result_model is None:
            return
        timers = result_model.timers
        if not isinstance(timers, list) or len(timers) == 0:
            self.state.timer = None
            return
        if len(timers) > 1:
            logger.debug(
                (
                    (
                        'Multiple timers found - %s, this method '
                        'will only return the most recent timer created.'
                    ),
                ),
                timers,
            )
        timer = timers[0]
        if not isinstance(timer, TimerModels.TimerItemV1):
            logger.warning('Invalid timer model - %s', timer)
            return
        self.state.timer = Timer(
            timer_duration=int(timer.counterTimer),
            id=int(timer.timerID),
            action=timer.action,
        )

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if not isinstance(action, str) or action not in [
            DeviceStatus.ON,
            DeviceStatus.OFF,
        ]:
            logger.error('Invalid action for timer - %s', action)
            return False
        update_dict = {
            'action': action,
            'counterTime': str(duration),
            'switchNo': self.sub_device_no,
        }
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1SetTime,
            update_dict=update_dict,
            method='addTimer',
            endpoint='timer/addTimer',
        )
        result_model = process_bypassv1_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV1SetTimer
        )
        if result_model is None:
            return False
        self.state.timer = Timer(duration, action, int(result_model.timerID))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set, nothing to clear, run get_timer().')
            return False
        if self.state.timer.time_remaining == 0:
            logger.debug('Timer already ended.')
            self.state.timer = None
            return True
        r_dict = await self.call_bypassv1_api(
            TimerModels.RequestV1ClearTimer,
            {'timerId': str(self.state.timer.id)},
            method='deleteTimer',
            endpoint='timer/deleteTimer',
        )
        r_dict = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r_dict is None:
            return False
        self.state.timer = None
        return True


class VeSyncOutletWHOGPlug(BypassV2Mixin, VeSyncOutlet):
    """VeSync WHOG smart plug.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize BSDGO1 smart plug class."""
        super().__init__(details, manager, feature_map)

    async def _bypass_v1_api_helper(
        self,
        request_model: type[RequestBypassV1],
        update_dict: dict | None = None,
        method: str = 'bypass',
    ) -> dict | None:
        """Send ByPass V1 API request.

        This uses the `_build_request` method to send API requests to the Bypass V1 API.

        Args:
            request_model (type[RequestBypassV1]): The request model to use.
            update_dict (dict): Additional keys to add on.
            method (str): The method to use in the outer body.

        Returns:
            bytes: The response from the API request.
        """
        keys = BypassV1Mixin.request_keys
        body = Helpers.get_defaultvalues_attributes(keys)
        body.update(Helpers.get_manager_attributes(self.manager, keys))
        body.update(Helpers.get_device_attributes(self, keys))
        body['method'] = method
        body.update(update_dict or {})
        model_instance = request_model.from_dict(body)
        url_path = '/cloud/v1/outlet/getELECConsumePerMonthLastYear'
        resp_dict, _ = await self.manager.async_call_api(
            url_path, 'post', model_instance, Helpers.req_header_bypass()
        )

        return resp_dict

    def _set_state(self, resp_model: BypassV2InnerResult) -> None:
        """Set the state of the WHOGPLUG outlet from the response model."""
        if not isinstance(resp_model, ResponseWHOGResult):
            logger.debug('Invalid response model for _set_state: %s', type(resp_model))
            return
        self.state.device_status = DeviceStatus.from_int(resp_model.enabled)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.voltage = resp_model.voltage
        self.state.power = resp_model.power
        self.state.energy = resp_model.energy
        self.state.current = resp_model.current
        self.state.voltageUpperThreshold = resp_model.highestVoltage
        self.state.protectionStatus = 'on' if resp_model.voltagePtStatus else 'off'

    async def get_details(self) -> None:
        r_dict = await self.call_bypassv2_api('getOutletStatus')

        resp_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, ResponseWHOGResult
        )
        if resp_model is None:
            logger.debug('Error getting %s details', self.device_name)
            self.state.connection_status = ConnectionStatus.OFFLINE
            return

        self._set_state(resp_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        payload_data = {
            'id': 0,
            'enabled': toggle,
        }
        r_dict = await self.call_bypassv2_api(
            payload_method='setSwitch',
            data=payload_data,
        )
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def _get_energy_history(self, history_interval: str | EnergyIntervals) -> None:
        """Get energy history for BSDGO1 outlet."""
        if history_interval not in self._energy_intervals:
            logger.error('Invalid energy history interval - %s', history_interval)
            return
        if not isinstance(history_interval, EnergyIntervals):
            history_interval = EnergyIntervals(history_interval)

        if history_interval == EnergyIntervals.YEAR:
            await self.get_yearly_energy()

        offset = ENERGY_HISTORY_OFFSET_WHOGPLUG[history_interval]
        current_ts = int(datetime.now().timestamp())
        start_ts = current_ts - offset
        payload_data = {
            'fromDay': start_ts,
            'toDay': current_ts,
        }
        r_dict = await self.call_bypassv2_api('getEnergyHistory', data=payload_data)
        result_model = process_bypassv2_result(
            self, logger, '_get_energy_history', r_dict, ResponseEnergyResult
        )
        if result_model is None:
            return
        match history_interval:
            case EnergyIntervals.WEEK:
                self.state.weekly_history = result_model
            case EnergyIntervals.MONTH:
                self.state.monthly_history = result_model
        logger.debug('Energy history for %s updated', self.device_name)

    async def get_yearly_energy(self) -> None:
        """Get yearly energy for WHOG outlet."""
        r_dict = await self._bypass_v1_api_helper(
            RequestWHOGYearlyEnergy, method='getELECConsumePerMonthLastYear'
        )
        r_dict = Helpers.process_dev_response(logger, 'get_yearly_energy', self, r_dict)
        if r_dict is None:
            return
        if not isinstance(r_dict.get('result', {}).get('ELECConsumeList'), list):
            logger.debug('Error in last year energy response for %s', self.device_name)
            return

        self.state.yearly_history = ResponseEnergyResult.from_dict(
            self._process_yearly_model(r_dict['result'])
        )
        logger.debug('Last year energy for %s updated', self.device_name)

    def _process_yearly_model(
        self, result_dict: dict[str, list[dict[str, str]]]
    ) -> dict[str, list[dict]]:
        """Process yearly WHOG energy model from response dict."""

        def end_of_month_utc_timestamp(year: int, month: int) -> int:
            # Last day number in the month
            last_day = calendar.monthrange(year, month)[1]
            # 23:59:59 on the last day, in UTC
            dt = datetime(year, month, last_day, 23, 59, 59, tzinfo=UTC)
            return int(dt.timestamp())

        out: dict[str, list[dict]] = {'energyInfos': []}
        for item in result_dict.get('ELECConsumeList', []):
            ym = item['month']
            year, month = map(int, ym.split('-'))
            ts = end_of_month_utc_timestamp(year, month)
            out['energyInfos'].append(
                {
                    'timestamp': ts,
                    'energyKWH': item['ELECConsume'],
                }
            )
        return out


class VeSyncBSDOGPlug(VeSyncOutletWHOGPlug):
    """VeSync BSDOG01/WYZYOG smart plugs.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.

    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize BSDOG smart plug class."""
        super().__init__(details, manager, feature_map)

    def _set_state(self, resp_model: BypassV2InnerResult) -> None:
        """Set the state of the BSDOG outlet from the response model."""
        if not isinstance(resp_model, ResponseBSDGO1OutletResult):
            logger.debug('Invalid response model for _set_state: %s', type(resp_model))
            return
        self.state.device_status = DeviceStatus.from_int(resp_model.powerSwitch_1)
        self.state.connection_status = ConnectionStatus.ONLINE
        self.state.voltage = resp_model.realTimeVoltage
        self.state.power = resp_model.realTimePower
        self.state.energy = resp_model.electricalEnergy
        self.state.voltageUpperThreshold = resp_model.voltageUpperThreshold
        self.state.protectionStatus = resp_model.protectionStatus
        self.state.currentUpperThreshold = resp_model.currentUpperThreshold

    async def get_details(self) -> None:
        payload_data = {
            'properties': [
                'powerSwitch_1',
                'realTimeVoltage',
                'realTimePower',
                'electricalEnergy',
                'protectionStatus',
                'voltageUpperThreshold',
                'currentUpperThreshold',
                'scheduleNum',
            ]
        }

        r_dict = await self.call_bypassv2_api('getProperty', data=payload_data)

        resp_model = process_bypassv2_result(
            self, logger, 'get_details', r_dict, ResponseBSDGO1OutletResult
        )
        if resp_model is None:
            logger.debug('Error getting %s details', self.device_name)
            self.state.connection_status = ConnectionStatus.OFFLINE
            return

        self._set_state(resp_model)

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        toggle_int = int(toggle)
        r_dict = await self.call_bypassv2_api(
            'setProperty',
            data={'powerSwitch_1': toggle_int},
            payload_update={'subDeviceNo': 0},
        )
        r = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if r is None:
            return False

        self.state.device_status = DeviceStatus.ON if toggle else DeviceStatus.OFF
        self.state.connection_status = ConnectionStatus.ONLINE
        return True


class VeSyncESW10USA(BypassV2Mixin, VeSyncOutlet):
    """VeSync ESW10 USA outlet.

    Note that this device does not support energy monitoring.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        device_status (str): Device status.
        connection_status (str): Connection status.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize ESW10 USA outlet."""
        super().__init__(details, manager, feature_map)

    async def get_details(self) -> None:
        payload_data = {
            'id': 0,
        }
        payload_method = 'getSwitch'
        r_dict = await self.call_bypassv2_api(payload_method, payload_data)
        result = process_bypassv2_result(
            self, logger, 'get_details', r_dict, ResultESW10Details
        )
        if result is None:
            logger.warning('Error getting %s details', self.device_name)
            self.state.connection_status = ConnectionStatus.OFFLINE
            return
        self.state.device_status = DeviceStatus.from_bool(result.enabled)
        self.state.connection_status = ConnectionStatus.ONLINE

    async def toggle_switch(self, toggle: bool | None = None) -> bool:
        if toggle is None:
            toggle = self.state.device_status != DeviceStatus.ON
        payload_data = {
            'id': 0,
            'enabled': toggle,
        }
        payload_method = 'setSwitch'
        r_dict = await self.call_bypassv2_api(payload_method, payload_data)
        result = Helpers.process_dev_response(logger, 'toggle_switch', self, r_dict)
        if not isinstance(result, dict):
            logger.warning('Error toggling %s switch', self.device_name)
            return False
        self.state.device_status = DeviceStatus.from_bool(toggle)
        self.state.connection_status = ConnectionStatus.ONLINE
        return True

    async def get_timer(self) -> Timer | None:
        r_dict = await self.call_bypassv2_api('getTimer')
        result_model = process_bypassv2_result(
            self, logger, 'get_timer', r_dict, TimerModels.ResultV2GetTimer
        )
        if result_model is None:
            return None
        timers = result_model.timers
        if not timers:
            self.state.timer = None
            return None
        if len(timers) > 1:
            logger.debug(
                (
                    'Multiple timers found - %s, this method '
                    'will only return the most recent timer created.'
                ),
                timers,
            )
        timer = timers[0]
        self.state.timer = Timer(
            timer_duration=int(timer.remain),
            id=int(timer.id),
            action=timer.action,
        )
        return self.state.timer

    async def set_timer(self, duration: int, action: str | None = None) -> bool:
        if action is None:
            action = (
                DeviceStatus.ON
                if self.state.device_status == DeviceStatus.OFF
                else DeviceStatus.OFF
            )
        if action not in [DeviceStatus.ON, DeviceStatus.OFF]:
            logger.error('Invalid action for timer - %s', action)
            return False
        payload_data = {'action': action, 'total': duration}
        r_dict = await self.call_bypassv2_api(
            payload_method='addTimer',
            data=payload_data,
        )
        result_model = process_bypassv2_result(
            self, logger, 'set_timer', r_dict, TimerModels.ResultV2SetTimer
        )
        if result_model is None:
            return False
        if result_model.id is None:
            logger.warning('Unable to set timer.')
            return False
        self.state.timer = Timer(duration, action, int(result_model.id))
        return True

    async def clear_timer(self) -> bool:
        if self.state.timer is None:
            logger.debug('No timer set, nothing to clear, run get_timer().')
            return False
        if self.state.timer.time_remaining == 0:
            logger.debug('Timer already ended.')
            self.state.timer = None
            return True
        r_dict = await self.call_bypassv2_api(
            payload_method='delTimer', data={'id': self.state.timer.id}
        )
        r_dict = Helpers.process_dev_response(logger, 'clear_timer', self, r_dict)
        if r_dict is None:
            return False
        self.state.timer = None
        return True



"""Data models for VeSync air fryers."""

from __future__ import annotations

from dataclasses import dataclass

from pyvesync.models.base_models import ResponseBaseModel


@dataclass
class ResultFryerDetails(ResponseBaseModel):
    """Result model for air fryer details."""

    returnStatus: FryerCookingReturnStatus | FryerBaseReturnStatus | None = None


@dataclass
class FryerCookingReturnStatus(ResponseBaseModel):
    """Result returnStatus model for air fryer status."""

    currentTemp: int
    cookSetTemp: int
    mode: str
    cookSetTime: int
    cookLastTime: int
    cookStatus: str
    tempUnit: str


@dataclass
class FryerBaseReturnStatus(ResponseBaseModel):
    """Result returnStatus model for air fryer status."""

    cookStatus: str



"""Tests for VeSyncAuth authentication module.

Tests cover:
- Initialization and properties (region mapping, credentials state)
- Credential management (set, clear, output)
- File-based credential persistence (save, load)
- Login flow (get_authorization_code -> exchange for token)
- Re-authentication
- Error handling (bad responses, missing fields, cross-region)
"""

import logging
from pathlib import Path

import orjson
import pytest

from pyvesync import VeSync
from pyvesync.utils.errors import (
    VeSyncAPIResponseError,
    VeSyncLoginError,
    VeSyncServerError,
)

from base_test_cases import TestBase
from defaults import TestDefaults
import call_json

logger = logging.getLogger(__name__)

LOGIN_RESPONSES = call_json.LoginResponses


# ---------------------------------------------------------------------------
# Initialization & Properties
# ---------------------------------------------------------------------------


class TestAuthInit(TestBase):
    """Test VeSyncAuth initialization and property accessors."""

    def test_default_region_us(self):
        """US country code maps to US region."""
        assert self.manager.auth._country_code == 'US'
        assert self.manager.auth.current_region == 'US'

    def test_region_eu(self):
        """Non-US/CA/MX/JP country codes map to EU region."""
        manager = VeSync('user', 'pass', country_code='DE')
        assert manager.auth.current_region == 'EU'

    def test_region_jp(self):
        """JP country code maps to US region."""
        manager = VeSync('user', 'pass', country_code='JP')
        assert manager.auth.current_region == 'US'

    def test_country_code_uppercased(self):
        """Country code is stored uppercased."""
        manager = VeSync('user', 'pass', country_code='ca')
        assert manager.auth.country_code == 'CA'

    def test_country_code_setter_uppercases(self):
        """Country code setter uppercases value."""
        self.manager.auth.country_code = 'mx'
        assert self.manager.auth.country_code == 'MX'

    def test_not_authenticated_initially(self):
        """Auth is not authenticated before login."""
        manager = VeSync('user', 'pass')
        assert manager.auth.is_authenticated is False

    def test_token_raises_before_login(self):
        """Accessing token before login raises AttributeError."""
        manager = VeSync('user', 'pass')
        with pytest.raises(AttributeError, match='Token not set'):
            _ = manager.auth.token

    def test_account_id_raises_before_login(self):
        """Accessing account_id before login raises AttributeError."""
        manager = VeSync('user', 'pass')
        with pytest.raises(AttributeError, match='Account ID not set'):
            _ = manager.auth.account_id

    def test_repr(self):
        """Test __repr__ output."""
        manager = VeSync('user@test.com', 'pass')
        r = repr(manager.auth)
        assert 'user@test.com' in r
        assert 'authenticated=False' in r


# ---------------------------------------------------------------------------
# Credential Management
# ---------------------------------------------------------------------------


class TestCredentialManagement(TestBase):
    """Test set_credentials, clear_credentials, and output methods."""

    def test_set_credentials(self):
        """set_credentials sets all fields and marks authenticated."""
        auth = self.manager.auth
        auth.set_credentials(
            token='tk_123',
            account_id='acct_456',
            country_code='de',
            region='EU',
        )
        assert auth.is_authenticated is True
        assert auth.token == 'tk_123'
        assert auth.account_id == 'acct_456'
        assert auth.country_code == 'DE'
        assert auth.current_region == 'EU'

    def test_clear_credentials(self):
        """clear_credentials resets token and account_id."""
        auth = self.manager.auth
        auth.set_credentials('tk', 'acct', 'US', 'US')
        assert auth.is_authenticated is True
        auth.clear_credentials()
        assert auth.is_authenticated is False

    def test_output_credentials_dict_authenticated(self):
        """output_credentials_dict returns dict when authenticated."""
        auth = self.manager.auth
        auth.set_credentials('tk_abc', 'acct_xyz', 'US', 'US')
        result = auth.output_credentials_dict()
        assert result == {
            'token': 'tk_abc',
            'account_id': 'acct_xyz',
            'country_code': 'US',
            'current_region': 'US',
        }

    def test_output_credentials_dict_not_authenticated(self):
        """output_credentials_dict returns None when not authenticated."""
        manager = VeSync('user', 'pass')
        assert manager.auth.output_credentials_dict() is None

    def test_output_credentials_json_authenticated(self):
        """output_credentials_json returns JSON string when authenticated."""
        auth = self.manager.auth
        auth.set_credentials('tk', 'acct', 'US', 'US')
        json_str = auth.output_credentials_json()
        assert json_str is not None
        parsed = orjson.loads(json_str)
        assert parsed['token'] == 'tk'
        assert parsed['account_id'] == 'acct'

    def test_output_credentials_json_not_authenticated(self):
        """output_credentials_json returns None when not authenticated."""
        manager = VeSync('user', 'pass')
        assert manager.auth.output_credentials_json() is None


# ---------------------------------------------------------------------------
# File-based Credential Persistence
# ---------------------------------------------------------------------------


class TestCredentialFile(TestBase):
    """Test save/load credentials to/from file."""

    def test_save_and_load_credentials(self, tmp_path):
        """Save credentials to file and load them back."""
        auth = self.manager.auth
        auth.set_credentials('tk_save', 'acct_save', 'US', 'US')

        file_path = tmp_path / '.vesync_auth'
        self.run_in_loop(auth.save_credentials_to_file, file_path)
        assert file_path.exists()

        # Load into a fresh manager
        manager2 = VeSync('user', 'pass')
        success = self.loop.run_until_complete(
            manager2.auth.load_credentials_from_file(file_path)
        )
        assert success is True
        assert manager2.auth.token == 'tk_save'
        assert manager2.auth.account_id == 'acct_save'
        assert manager2.auth.country_code == 'US'
        assert manager2.enabled is True

    def test_load_nonexistent_file(self, tmp_path):
        """Loading from a nonexistent file returns False."""
        auth = self.manager.auth
        result = self.loop.run_until_complete(
            auth.load_credentials_from_file(tmp_path / 'missing_file')
        )
        assert result is False

    def test_load_invalid_json(self, tmp_path):
        """Loading a file with invalid JSON returns False."""
        file_path = tmp_path / '.vesync_auth'
        file_path.write_text('not json', encoding='utf-8')

        auth = self.manager.auth
        result = self.loop.run_until_complete(
            auth.load_credentials_from_file(file_path)
        )
        assert result is False

    def test_load_missing_keys(self, tmp_path):
        """Loading a file with missing keys returns False."""
        file_path = tmp_path / '.vesync_auth'
        file_path.write_text(
            orjson.dumps({'token': 'tk'}).decode('utf-8'),
            encoding='utf-8',
        )
        auth = self.manager.auth
        result = self.loop.run_until_complete(
            auth.load_credentials_from_file(file_path)
        )
        assert result is False

    def test_save_not_authenticated(self, tmp_path):
        """Saving credentials when not authenticated does nothing."""
        manager = VeSync('user', 'pass')
        file_path = tmp_path / '.vesync_auth'
        self.loop.run_until_complete(
            manager.auth.save_credentials_to_file(file_path)
        )
        assert not file_path.exists()

    def test_clear_credentials_deletes_file(self, tmp_path):
        """clear_credentials deletes the token file."""
        auth = self.manager.auth
        auth.set_credentials('tk', 'acct', 'US', 'US')

        file_path = tmp_path / '.vesync_auth'
        self.loop.run_until_complete(auth.save_credentials_to_file(file_path))
        assert file_path.exists()

        # Set the token file path (load sets this, but we saved directly)
        auth._token_file_path = file_path
        auth.clear_credentials()
        assert not file_path.exists()

    def test_credentials_saved_property(self, tmp_path):
        """credentials_saved reflects whether a token file exists."""
        auth = self.manager.auth
        assert auth.credentials_saved is False

        auth.set_credentials('tk', 'acct', 'US', 'US')
        file_path = tmp_path / '.vesync_auth'
        self.loop.run_until_complete(auth.save_credentials_to_file(file_path))
        auth._token_file_path = file_path
        assert auth.credentials_saved is True

    def test_load_with_path_object(self, tmp_path):
        """Loading credentials accepts a Path object."""
        auth = self.manager.auth
        auth.set_credentials('tk_path', 'acct_path', 'CA', 'US')

        file_path = tmp_path / '.vesync_auth'
        self.loop.run_until_complete(auth.save_credentials_to_file(file_path))

        manager2 = VeSync('user', 'pass')
        success = self.loop.run_until_complete(
            manager2.auth.load_credentials_from_file(Path(file_path))
        )
        assert success is True
        assert manager2.auth.token == 'tk_path'


# ---------------------------------------------------------------------------
# Login Flow
# ---------------------------------------------------------------------------


class TestLoginFlow(TestBase):
    """Test login flow with mocked async_call_api."""

    def test_login_success(self):
        """Successful two-step login sets token and account_id."""
        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS, 200),
        ]
        result = self.run_in_loop(self.manager.auth.login)
        assert result is True
        assert self.manager.auth.is_authenticated is True
        assert self.manager.auth.token == TestDefaults.token
        assert self.manager.auth.account_id == TestDefaults.account_id

    def test_login_no_credentials(self):
        """Login without username/password raises VeSyncLoginError."""
        manager = VeSync('', '')
        manager.auth._token = None
        manager.auth._account_id = None
        with pytest.raises(VeSyncLoginError, match='No valid authentication'):
            self.loop.run_until_complete(manager.auth.login())

    def test_login_auth_error_code(self):
        """Non-zero code in auth response raises VeSyncLoginError."""
        error_resp = {
            'traceId': TestDefaults.trace_id,
            'code': -11201000,
            'msg': 'Invalid password',
            'result': None,
        }
        self.mock_api.return_value = (error_resp, 200)
        with pytest.raises(VeSyncLoginError, match='Authentication failed'):
            self.run_in_loop(self.manager.auth.login)

    def test_login_none_response(self):
        """None API response raises VeSyncAPIResponseError."""
        self.mock_api.return_value = (None, 200)
        with pytest.raises(
            VeSyncAPIResponseError, match='Error receiving response'
        ):
            self.run_in_loop(self.manager.auth.login)

    def test_login_exchange_error_code(self):
        """Non-zero code in exchange response raises VeSyncLoginError."""
        login_error_resp = {
            'traceId': TestDefaults.trace_id,
            'code': -11201000,
            'msg': 'Login failed',
            'result': {
                'accountID': TestDefaults.account_id,
                'acceptLanguage': 'en',
                'countryCode': 'US',
                'token': '',
            },
        }
        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (login_error_resp, 200),
        ]
        with pytest.raises(VeSyncLoginError, match='Login failed'):
            self.run_in_loop(self.manager.auth.login)

    def test_login_exchange_none_response(self):
        """None response in exchange step raises VeSyncAPIResponseError."""
        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (None, 200),
        ]
        with pytest.raises(
            VeSyncAPIResponseError, match='Error receiving response'
        ):
            self.run_in_loop(self.manager.auth.login)

    def test_login_cross_region(self):
        """Cross-region error triggers retry with new region."""
        cross_region_resp = LOGIN_RESPONSES.LOGIN_RESPONSE_CROSS_REGION
        success_resp = LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS
        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (cross_region_resp, 200),
            (success_resp, 200),
        ]
        result = self.run_in_loop(self.manager.auth.login)
        assert result is True
        assert self.mock_api.call_count == 3

    def test_login_server_error_in_exchange(self):
        """Server error code in exchange raises VeSyncServerError."""
        server_error_resp = {
            'traceId': TestDefaults.trace_id,
            'code': -11102000,
            'msg': 'Server error',
            'result': {
                'accountID': TestDefaults.account_id,
                'acceptLanguage': 'en',
                'countryCode': 'US',
                'token': '',
            },
        }
        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (server_error_resp, 200),
        ]
        with pytest.raises(VeSyncServerError):
            self.run_in_loop(self.manager.auth.login)


# ---------------------------------------------------------------------------
# Re-authentication
# ---------------------------------------------------------------------------


class TestReauthentication(TestBase):
    """Test reauthenticate method."""

    def test_reauthenticate_success(self):
        """Reauthenticate clears and re-establishes credentials."""
        auth = self.manager.auth
        auth.set_credentials('old_tk', 'old_acct', 'US', 'US')
        assert auth.is_authenticated is True

        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS, 200),
        ]
        result = self.run_in_loop(auth.reauthenticate)
        assert result is True
        assert auth.token == TestDefaults.token

    def test_reauthenticate_saves_to_file(self, tmp_path):
        """Reauthenticate re-saves credentials when file existed."""
        auth = self.manager.auth
        auth.set_credentials('tk', 'acct', 'US', 'US')
        file_path = tmp_path / '.vesync_auth'
        self.loop.run_until_complete(auth.save_credentials_to_file(file_path))
        auth._token_file_path = file_path

        self.mock_api.side_effect = [
            (LOGIN_RESPONSES.GET_TOKEN_RESPONSE_SUCCESS, 200),
            (LOGIN_RESPONSES.LOGIN_RESPONSE_SUCCESS, 200),
        ]
        self.run_in_loop(auth.reauthenticate)

        # File should be re-created with new credentials
        assert file_path.exists()
        data = orjson.loads(file_path.read_text(encoding='utf-8'))
        assert data['token'] == TestDefaults.token

    def test_reauthenticate_login_failure_raises(self):
        """Reauthenticate propagates login errors."""
        auth = self.manager.auth
        auth.set_credentials('tk', 'acct', 'US', 'US')

        self.mock_api.return_value = (None, 200)
        with pytest.raises(VeSyncAPIResponseError):
            self.run_in_loop(auth.reauthenticate)
        assert auth.is_authenticated is False



"""General VeSync tests."""

import asyncio
import orjson
import pytest
from unittest.mock import patch, MagicMock

from pyvesync import VeSync
from pyvesync.utils.errors import VeSyncRateLimitError, VeSyncServerError

from pyvesync.const import API_BASE_URL_US
from pyvesync.utils.errors import (
    VeSyncAPIStatusCodeError
    )
import call_json
from defaults import TestDefaults
from aiohttp_mocker import AiohttpMockSession


DEFAULT_ENDPOINT = '/endpoint'
DEFAULT_POST_DATA = {'key': 'value'}
PARAM_ARGS = "endpoint, method, resp_bytes, resp_status"

# Successful API calls should return the response in bytes and a 200 status code
SUCCESS_RESP = call_json.response_body(0, 'Success')


# Rate limit errors should raise an exception in `async_call_api`
RATE_LIMIT_CODE = -11003000
RATE_LIMIT_RESP = call_json.response_body(RATE_LIMIT_CODE, "Rate limit exceeded")


# Server errors should raise an exception in `async_call_api`
SERVER_ERROR = -11102000
SERVER_ERROR_RESP = call_json.response_body(SERVER_ERROR, "Server error")


# Status code errors should raise an exception in `async_call_api`
STATUS_CODE_ERROR = 400
STATUS_CODE_RESP = None


# Login errors should not raise an exception in `async_call_api`, but are raised in login() method
STATUS_CODE_ERROR = 400
STATUS_CODE_RESP = None


# Device errors should return the response and a 200 status code
# with no exception thrown by `async_call_api`
DEVICE_ERROR_CODE = -11901000
DEVICE_ERROR_RESP = call_json.response_body(DEVICE_ERROR_CODE, "Device error")


class TestApiFunc:
    """Test call_api() method."""
    @pytest.fixture(autouse=True, scope='function')
    def setup(self, caplog):
        """Fixture to instantiate VeSync object, start logging and start Mock.

        Attributes
        ----------
        self.mock_api : Mock
        self.manager : VeSync
        self.caplog : LogCaptureFixture

        Yields
        ------
        Class instance with mocked call_api() function and VeSync object
        """
        self.caplog = caplog
        self.loop = asyncio.new_event_loop()
        self.mock = MagicMock()
        self.manager = VeSync('EMAIL', 'PASSWORD')
        self.manager.enabled = True
        self.manager.auth._token = TestDefaults.token
        self.manager.auth._account_id = TestDefaults.account_id
        yield
        self.mock.stop()
        self.loop.stop()

    async def run_coro(self, coro):
        """Run a coroutine in the event loop."""
        return await coro

    def run_in_loop(self, func, *args, **kwargs):
        """Run a function in the event loop."""
        return self.loop.run_until_complete(self.run_coro(func(*args, **kwargs)))

    @patch("pyvesync.vesync.ClientSession")
    def test_api_success(self, mock):
        """Test successful api call - returns tuple of response bytes and status code."""
        mock.return_value.request.return_value = AiohttpMockSession(
            method='post',
            url=API_BASE_URL_US + DEFAULT_ENDPOINT,
            status=200,
            response=orjson.dumps(SUCCESS_RESP),
        )
        resp = self.run_in_loop(self.manager.async_call_api, DEFAULT_ENDPOINT, method="get")
        mock_dict, mock_status = resp

        assert SUCCESS_RESP == mock_dict
        assert mock_status == 200

    @patch("pyvesync.vesync.ClientSession")
    def test_api_rate_limit(self, mock):
        """Test rate limit error - raises `VeSyncRateLimitError` from `VeSync.async_call_api`."""
        rate_limit_resp = call_json.response_body(RATE_LIMIT_CODE, "Rate limit exceeded")
        mock.return_value.request.return_value = AiohttpMockSession(
            method="post",
            url=API_BASE_URL_US + DEFAULT_ENDPOINT,
            status=200,
            response=orjson.dumps(rate_limit_resp),
        )
        with pytest.raises(VeSyncRateLimitError):
            self.run_in_loop(self.manager.async_call_api, DEFAULT_ENDPOINT, 'post', json_object=DEFAULT_POST_DATA)

    @patch("pyvesync.vesync.ClientSession")
    def test_api_server_error(self, mock):
        """Test server error - raises `VeSyncServerError` from `VeSync.async_call_api`."""
        mock.return_value.request.return_value = AiohttpMockSession(
            method='post',
            url=API_BASE_URL_US + DEFAULT_ENDPOINT,
            status=200,
            response=orjson.dumps(SERVER_ERROR_RESP),
        )
        with pytest.raises(VeSyncServerError):
            self.run_in_loop(self.manager.async_call_api, DEFAULT_ENDPOINT, 'post', json_object=DEFAULT_POST_DATA)

    @patch("pyvesync.vesync.ClientSession")
    def test_api_status_code_error(self, mock):
        """Test status code error - raises `VeSyncAPIStatusCodeError` from `VeSync.async_call_api`."""
        mock.return_value.request.return_value = AiohttpMockSession(
            method='get',
            url=API_BASE_URL_US + DEFAULT_ENDPOINT,
            status=404,
            response=None,
        )
        with pytest.raises(VeSyncAPIStatusCodeError):
            self.run_in_loop(self.manager.async_call_api, DEFAULT_ENDPOINT, 'get')



import orjson
import pytest
from unittest.mock import patch
import logging
from pyvesync import VeSync
from pyvesync.devices.vesyncswitch import VeSyncWallSwitch
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import call_json_switches
from base_test_cases import TestBase

DEVICE_TYPE = 'ESWL01'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_TYPE)

CORRECT_WS_LIST = call_json.DeviceList.device_list_response(DEVICE_TYPE)

CORRECT_WS_DETAILS = call_json_switches.DETAILS_RESPONSES[DEVICE_TYPE]

BAD_LIST = call_json.DETAILS_BADCODE

DEFAULTS = call_json.TestDefaults


class TestVesyncWallSwitch(TestBase):
    # @pytest.fixture()
    # def api_mock(self, caplog):
    #     self.mock_api_call = patch('pyvesync.vesync.VeSync.async_call_api')
    #     self.mock_api = self.mock_api_call.start()
    #     self.mock_api.create_autospect()
    #     self.mock_api.return_value.ok = True
    #     self.vesync_obj = VeSync('sam@mail.com', 'pass', debug=True)
    #     self.vesync_obj.enabled = True
    #     self.vesync_obj.login = True
    #     self.vesync_obj.token = DEFAULTS.token
    #     self.vesync_obj.account_id = DEFAULTS.account_id
    #     caplog.set_level(logging.DEBUG)
    #     yield
    #     self.mock_api_call.stop()

    def test_ws_conf(self):
        """Tests that Wall Switch is instantiated properly"""
        self.mock_api.return_value = CORRECT_WS_LIST
        self.run_in_loop(self.manager.get_devices)
        switch = self.manager.switches
        assert len(switch) == 1
        wswitch = switch[0]
        assert isinstance(wswitch, VeSyncWallSwitch)
        assert wswitch.device_name == DEFAULTS.name(DEVICE_TYPE)
        assert wswitch.device_type == DEVICE_TYPE
        assert wswitch.cid == DEFAULTS.cid(DEVICE_TYPE)
        assert wswitch.uuid == DEFAULTS.uuid(DEVICE_TYPE)

    def test_ws_details(self):
        """Test WS get_details() """
        resp_dict, status = CORRECT_WS_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        wswitch = VeSyncWallSwitch(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(wswitch.get_details)
        dev_details = wswitch.details
        assert wswitch.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert dev_details['active_time'] == 1
        assert wswitch.connection_status == 'online'

    def test_ws_details_fail(self, caplog):
        """Test WS get_details with Code>0"""
        resp_dict, status = BAD_LIST
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch15a = VeSyncWallSwitch(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_details)
        assert len(caplog.records) == 1
        assert 'details' in caplog.text

    def test_ws_onoff(self, caplog):
        """Test 15A Device On/Off Methods"""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        wswitch = VeSyncWallSwitch(DEV_LIST_DETAIL, self.manager)
        head = Helpers.req_legacy_headers(self.manager)
        body = Helpers.req_body(self.manager, 'devicestatus')

        body['status'] = 'on'
        body['uuid'] = wswitch.uuid
        on = self.run_in_loop(wswitch.turn_on)
        self.mock_api.assert_called_with(
            '/inwallswitch/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert on
        off = self.run_in_loop(wswitch.turn_off)
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/inwallswitch/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert off

    def test_ws_onoff_fail(self):
        """Test ws On/Off Fail with Code>0"""
        self.mock_api.return_value = (orjson.dumps({'code': 1}), 400)
        vswitch15a = VeSyncWallSwitch(DEV_LIST_DETAIL, self.manager)
        assert not self.run_in_loop(vswitch15a.turn_on)
        assert not self.run_in_loop(vswitch15a.turn_off)



"""
This tests requests made by bulb devices.

All tests inherit from the TestBase class which contains the fixtures
and methods needed to run the tests.

The tests are automatically parametrized by `pytest_generate_tests` in
conftest.py. The two methods that are parametrized are `test_details`
and `test_methods`. The class variables are used to build the list of
devices, test methods and arguments.

The `helpers.call_api` method is patched to return a mock response.
The method, endpoint, headers and json arguments are recorded
in YAML files in the api directory, categorized in folders by
module and files by the class name.

The default is to record requests that do not exist and compare requests
that already exist. If the API changes, set the overwrite argument to True
in order to overwrite the existing YAML file with the new request.

See Also
--------
`utils.TestBase` - Base class for all tests, containing mock objects
`confest.pytest_generate_tests` - Parametrizes tests based on
    method names & class attributes
`call_json_bulbs` - Contains API responses
"""

import logging
import math
from dataclasses import asdict
import pyvesync.const as const
from pyvesync.utils.helpers import Converters
from pyvesync.base_devices.bulb_base import VeSyncBulb
from base_test_cases import TestBase
from utils import assert_test, parse_args
from defaults import TestDefaults
import call_json_bulbs


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

DEFAULT_COLOR = TestDefaults.color
DEFAULT_COLOR_RGB = dict(asdict(DEFAULT_COLOR.rgb))
DEFAULT_COLOR_HSV = dict(asdict(DEFAULT_COLOR.hsv))
RGB_SET = {
    'red': 50,
    'green': 200,
    'blue': 255,

}
HSV_SET = {
    'hue': 200,
    'saturation': 50,
    'value': 100,
}


class TestBulbs(TestBase):
    """Bulbs testing class.

    This class tests bulb device details and methods. The methods are
    parametrized from the class variables using `pytest_generate_tests`.
    The call_json_bulbs module contains the responses for the API requests.
    The device is instantiated from the details provided by
    `call_json.DeviceList.device_list_item()`. Inherits from `utils.TestBase`.

    Instance Attributes
    -------------------
    self.manager : VeSync
        Instantiated VeSync object
    self.mock_api : Mock
        Mock with patched `helpers.call_api` method
    self.caplog : LogCaptureFixture
        Pytest fixture for capturing logs

    Class Attributes
    -----------------
    device : str
        Name of product class - bulbs
    bulbs : list
        List of setup_entry's for bulbs, this variable is named
        after the device class attribute value
    base_methods : List[List[str, Dict[str, Any]]]
        List of common methods for all devices
    device_methods : Dict[List[List[str, Dict[str, Any]]]]
        Dictionary of methods specific to setup_entry's for each
        device.

    Methods
    --------
    test_details()
        Test the device details API request and response
    test_methods()
        Test device methods API request and response

    Examples
    --------
    >>> device = 'bulbs'
    >>> bulbs = call_json_bulbs.bulbs
    >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
    >>> device_methods = {
        'ESWD16': [['method1'], ['method2', {'kwargs': 'value'}]]
        }

    """

    device = 'bulbs'
    bulbs = call_json_bulbs.BULBS
    base_methods = [['turn_on'], ['turn_off'], ['set_brightness', {'brightness': 50}]]
    device_methods = {
        'ESL100CW': [['set_color_temp', {'color_temp': 50}]],
        'ESL100MC': [['set_rgb', RGB_SET],
                     ['set_white_mode']],
        'XYD0001': [['set_hsv', HSV_SET],
                    ['set_color_temp', {'color_temp': 50}],
                    ['set_white_mode']
                    ]
    }

    def test_details(self, setup_entry, method):
        """Test the device details API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class attribute `device` (name of product class - bulbs),
        device name (bulbs) list of `setup_entry` strings for each object.

        Example:
            >>> device = 'bulbs'
            >>> bulbs = call_json_bulbs.BULBS

        See Also
        --------
        `utils.TestBase` class docstring
        `call_json_bulbs` module docstring

        Notes
        ------
        The device is instantiated using the `call_json.DeviceList.device_list_item()`
        method. The device details contain the default values set in `utils.Defaults`
        """
        # Set return value for call_api based on call_json_bulb.DETAILS_RESPONSES
        return_dict = call_json_bulbs.DETAILS_RESPONSES[setup_entry]
        return_val = (return_dict, 200)
        self.mock_api.return_value = return_val

        # Instantiate device from device list return item
        bulb_obj = self.get_device("bulbs", setup_entry)
        assert isinstance(bulb_obj, VeSyncBulb)

        method_call = getattr(bulb_obj, method)
        self.run_in_loop(method_call)

        # Parse mock_api args tuple from arg, kwargs to kwargs
        all_kwargs = parse_args(self.mock_api)

        # Assert request matches recorded request or write new records
        assert assert_test(
            method_call, all_kwargs, setup_entry, self.write_api, self.overwrite
        )

        # Assert device details match expected values
        assert bulb_obj.state.brightness == TestDefaults.brightness
        if bulb_obj.supports_multicolor:
            assert self._assert_color(bulb_obj)
        if bulb_obj.supports_color_temp:
            assert bulb_obj.state.color_temp == TestDefaults.color_temp
            assert bulb_obj.state.color_temp_kelvin == Converters.color_temp_pct_to_kelvin(TestDefaults.color_temp)

    def test_methods(self, setup_entry, method):
        """Test device methods API request and response.

        This method is automatically parametrized by `pytest_generate_tests`
        based on class variables `device` (name of product class - bulbs),
        device name (bulbs) list of `setup_entry`'s, `base_methods` - list of
        methods for all devices, and `device_methods` - list of methods for
        each setup_entry.

        Example:
            >>> base_methods = [['turn_on'], ['turn_off'], ['update']]
            >>> device_methods = {
                'setup_entry': [['method1'], ['method2', {'kwargs': 'value'}]]
                }

        Notes
        -----
        The response can be a callable that accepts the `kwargs` argument to
        sync the device response with the API response. In some cases the API
        returns data from the method call, such as `get_yearly_energy`, in other cases the
        API returns a simple confirmation the command was successful.

        See Also
        --------
        `TestBase` class method
        `call_json_bulbs` module

        """
        # Get method name and kwargs from method fixture
        method_name = method[0]
        if len(method) == 2 and isinstance(method[1], dict):
            method_kwargs = method[1]
        else:
            method_kwargs = {}

        # Set return value for call_api based on call_json_bulbs.METHOD_RESPONSES
        method_response = call_json_bulbs.METHOD_RESPONSES[setup_entry][method_name]
        if callable(method_response):
            if method_kwargs:
                self.mock_api.return_value = method_response(method_kwargs), 200
            else:
                self.mock_api.return_value = method_response(), 200
        else:
            self.mock_api.return_value = method_response, 200

        # Get device configuration from call_json.DeviceList.device_list_item()

        bulb_obj = self.get_device("bulbs", setup_entry)
        assert isinstance(bulb_obj, VeSyncBulb)

        # Get method from device object
        method_call = getattr(bulb_obj, method[0])

        # Ensure method runs based on device configuration
        if method[0] == 'turn_on':
            bulb_obj.state.device_status = const.DeviceStatus.OFF
        elif method[0] == 'turn_off':
            bulb_obj.state.device_status = const.DeviceStatus.ON

        # Call method with kwargs if defined
        if method_kwargs:
            self.run_in_loop(method_call, **method_kwargs)
        else:
            self.run_in_loop(method_call)

        # Parse arguments from mock_api call into a dictionary
        all_kwargs = parse_args(self.mock_api)

        # Assert request matches recorded request or write new records
        assert assert_test(
            method_call, all_kwargs, setup_entry, self.write_api, self.overwrite
        )

    def _assert_color(self, bulb_obj):
        assert math.isclose(bulb_obj.state.color.rgb.red, DEFAULT_COLOR.rgb.red, rel_tol=1)
        assert math.isclose(bulb_obj.state.color.rgb.green, DEFAULT_COLOR.rgb.green, rel_tol=1)
        assert math.isclose(bulb_obj.state.color.rgb.blue, DEFAULT_COLOR.rgb.blue, rel_tol=1)
        assert math.isclose(bulb_obj.state.color.hsv.hue, DEFAULT_COLOR.hsv.hue, rel_tol=1)
        assert math.isclose(bulb_obj.state.color.hsv.saturation,
                            DEFAULT_COLOR.hsv.saturation,
                            rel_tol=1)
        assert math.isclose(bulb_obj.state.color.hsv.value, DEFAULT_COLOR.hsv.value, rel_tol=1)
        return True



"""Base classes for thermostat devices."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseDevice
from pyvesync.const import (
    ThermostatConst,
    ThermostatEcoTypes,
    ThermostatFanModes,
    ThermostatFanStatus,
    ThermostatHoldOptions,
    ThermostatScheduleOrHoldOptions,
    ThermostatWorkModes,
    ThermostatWorkStatusCodes,
)

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import ThermostatMap
    from pyvesync.models.thermostat_models import (
        ThermostatMinorDetails,
        ThermostatSimpleRoutine,
    )
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


_LOGGER = logging.getLogger(__name__)


class ThermostatState(DeviceState):
    """VeSync Thermostat State.

    Args:
        device (VeSyncThermostat): The thermostat device.
        details (ResponseDeviceDetailsModel): The thermostat device details.
        feature_map (ThermostatMap): The thermostat feature map.
    """

    __slots__ = (
        'alert_status',
        'battery_level',
        'configuration',
        'cool_to_temp',
        'deadband',
        'device_config',
        'eco_type',
        'fan_mode',
        'fan_status',
        'filter_life',
        'heat_to_temp',
        'hold_end_time',
        'hold_option',
        'humidity',
        'lock_status',
        'mode',
        'routine_running_id',
        'routines',
        'schedule_or_hold',
        'temperature',
        'temperature_unit',
        'work_mode',
        'work_status',
    )

    def __init__(
        self,
        device: VeSyncThermostat,
        details: ResponseDeviceDetailsModel,
        feature_map: ThermostatMap,
    ) -> None:
        """Initialize VeSync Thermostat State."""
        super().__init__(device, details, feature_map)
        self.device: VeSyncThermostat = device
        self.configuration: ThermostatMinorDetails | None = None
        self.work_mode: ThermostatWorkModes | None = None
        self.work_status: ThermostatWorkStatusCodes | None = None
        self.fan_mode: ThermostatFanModes | None = None
        self.fan_status: ThermostatFanStatus | None = None
        self.temperature_unit: str | None = None
        self.temperature: float | None = None
        self.humidity: int | None = None
        self.heat_to_temp: int | None = None
        self.cool_to_temp: int | None = None
        self.lock_status: bool = False
        self.schedule_or_hold: ThermostatScheduleOrHoldOptions | None = None
        self.hold_end_time: int | None = None
        self.hold_option: ThermostatHoldOptions | None = None
        self.deadband: int | None = None
        self.eco_type: ThermostatEcoTypes | None = None
        self.alert_status: int | None = None
        self.routines: list[ThermostatSimpleRoutine] = []
        self.routine_running_id: int | None = None
        self.battery_level: int | None = None
        self.filter_life: int | None = None

    @property
    def is_running(self) -> bool:
        """Check if the thermostat is running."""
        return self.work_status != ThermostatWorkStatusCodes.OFF

    @property
    def is_heating(self) -> bool:
        """Check if the thermostat is heating."""
        return self.work_status == ThermostatWorkStatusCodes.HEATING

    @property
    def is_cooling(self) -> bool:
        """Check if the thermostat is cooling."""
        return self.work_status == ThermostatWorkStatusCodes.COOLING

    @property
    def is_fan_on(self) -> bool:
        """Check if the fan is on."""
        return self.fan_status == ThermostatFanStatus.ON


class VeSyncThermostat(VeSyncBaseDevice):
    """Base class for VeSync Thermostat devices.

    Args:
        details (ResponseDeviceDetailsModel): The thermostat device details.
        manager (VeSync): The VeSync manager instance.
        feature_map (ThermostatMap): The thermostat feature map.
    """

    __slots__ = ('eco_types', 'fan_modes', 'hold_options', 'supported_work_modes')

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: ThermostatMap,
    ) -> None:
        """Initialize VeSync Thermostat."""
        super().__init__(details, manager, feature_map)
        self.state: ThermostatState = ThermostatState(self, details, feature_map)
        self.fan_modes = feature_map.fan_modes
        self.supported_work_modes = feature_map.modes
        self.eco_types = feature_map.eco_types
        self.hold_options = feature_map.hold_options

    async def set_mode(self, mode: ThermostatWorkModes) -> bool:
        """Set the thermostat mode."""
        del mode  # Unused
        _LOGGER.debug('set mode not implemented for %s', self.device_type)
        return False

    async def turn_off(self) -> bool:
        """Set mode to off."""
        return await self.set_mode(ThermostatWorkModes.OFF)

    async def set_mode_cool(self) -> bool:
        """Set mode to cool."""
        return await self.set_mode(ThermostatWorkModes.COOL)

    async def set_mode_heat(self) -> bool:
        """Set mode to heat."""
        return await self.set_mode(ThermostatWorkModes.HEAT)

    async def set_mode_auto(self) -> bool:
        """Set mode to auto."""
        return await self.set_mode(ThermostatWorkModes.AUTO)

    async def set_mode_smart_auto(self) -> bool:
        """Set mode to smart auto."""
        return await self.set_mode(ThermostatWorkModes.SMART_AUTO)

    async def set_mode_emergency_heat(self) -> bool:
        """Set mode to emergency heat."""
        return await self.set_mode(ThermostatWorkModes.EM_HEAT)

    async def set_fan_mode(self, mode: ThermostatFanModes) -> bool:
        """Set thermostat fan mode."""
        del mode
        _LOGGER.debug('set fan mode not implemented for %s', self.device_type)
        return False

    async def set_fan_ciruclate(self) -> bool:
        """Set fan circulate."""
        return await self.set_fan_mode(ThermostatConst.FanMode.CIRCULATE)

    async def set_fan_auto(self) -> bool:
        """Set fan auto."""
        return await self.set_fan_mode(ThermostatConst.FanMode.AUTO)

    async def set_fan_on(self) -> bool:
        """Set fan on."""
        return await self.set_fan_mode(ThermostatConst.FanMode.ON)

    async def get_configuration(self) -> None:
        """Retrieve configuration from API."""
        _LOGGER.debug('get configuration not implemented for %s', self.device_type)

    async def set_temp_point(self, temperature: float) -> bool:
        """Set the temperature point."""
        del temperature
        _LOGGER.debug('set temp point not implemented for %s', self.device_type)
        return False

    async def cancel_hold(self) -> bool:
        """Cancel the scheduled hold."""
        _LOGGER.debug('cancel hold not implemented for %s', self.device_type)
        return False

    async def set_cool_to_temp(self, temperature: float) -> bool:
        """Set the cool to temperature.

        Args:
            temperature (float): The cool to temperature.

        Returns:
            bool: True if successful, False otherwise.
        """
        del temperature
        _LOGGER.debug('set cool to temp not implemented for %s', self.device_type)
        return False

    async def set_heat_to_temp(self, temperature: float) -> bool:
        """Set the heat to temperature.

        Args:
            temperature (float): The heat to temperature.

        Returns:
            bool: True if successful, False otherwise.
        """
        del temperature
        _LOGGER.debug('set heat to temp not implemented for %s', self.device_type)
        return False

    async def toggle_lock(self, toggle: bool, pin: int | str | None = None) -> bool:
        """Toggle the thermostat lock status."""
        del toggle, pin
        _LOGGER.debug('toggle lock not implemented for %s', self.device_type)
        return False

    async def turn_on_lock(self, pin: int | str) -> bool:
        """Turn on the thermostat lock.

        Args:
            pin (int | str): The 4-digit PIN code.
        """
        return await self.toggle_lock(True, pin)

    async def turn_off_lock(self) -> bool:
        """Turn off the thermostat lock."""
        return await self.toggle_lock(False)

    async def set_eco_type(self, eco_type: ThermostatEcoTypes) -> bool:
        """Set thermostat eco type.

        Args:
            eco_type (ThermostatEcoTypes): The eco type to set, options are found in
                self.eco_types.

        Returns:
            bool: True if successful, False otherwise.
        """
        del eco_type
        if not self.eco_types:
            _LOGGER.debug('No eco types available for %s', self.device_type)
        else:
            _LOGGER.debug('set_eco_type not configured for %s', self.device_name)
        return False

    async def set_eco_first(self) -> bool:
        """Set eco first."""
        return await self.set_eco_type(ThermostatEcoTypes.ECO_FIRST)

    async def set_eco_second(self) -> bool:
        """Set eco second."""
        return await self.set_eco_type(ThermostatEcoTypes.ECO_SECOND)

    async def set_eco_comfort_first(self) -> bool:
        """Set eco comfort."""
        return await self.set_eco_type(ThermostatEcoTypes.COMFORT_FIRST)

    async def set_eco_comfort_second(self) -> bool:
        """Set eco comfort."""
        return await self.set_eco_type(ThermostatEcoTypes.COMFORT_SECOND)



"""pyvesync library constants.

All device states and information are defined by Enums in this module.

Attributes:
    DEFAULT_LANGUAGE (str): Default language for the VeSync app.
    API_BASE_URL (str): Base URL for the VeSync API. If not specified,
        a region-specific API URL is automatically selected.
    API_TIMEOUT (int): Timeout for API requests.
    USER_AGENT (str): User agent for API requests.
    DEFAULT_TZ (str): Default timezone for VeSync devices, updated by API after
        login.
    DEFAULT_REGION (str): Default region for VeSync devices,
        updated by API when retrieving devices.
    APP_VERSION (str): Version of the VeSync app.
    APP_ID (str): ID of the app. VeSync uses a random 8-letter string, but any non-empty
        string works.
    PHONE_BRAND (str): Brand of the phone used to login to the VeSync app.
    PHONE_OS (str): Operating system of the phone used to login to the VeSync app.
    MOBILE_ID (str): Unique identifier for the phone used to login to the VeSync app.
    USER_TYPE (str): User type for the VeSync app - internal app usage.
    BYPASS_APP_V (str): Bypass app version
    BYPASS_HEADER_UA (str): Bypass header user agent
    TERMINAL_ID (str): Unique identifier for new API calls
"""

from __future__ import annotations

import platform
import uuid
from enum import Enum, IntEnum, StrEnum
from random import randint
from types import MappingProxyType

from pyvesync.utils.enum_utils import IntEnumMixin

MAX_API_REAUTH_RETRIES = 3
DEFAULT_LANGUAGE = 'en'
API_BASE_URL = None  # Global URL (non-EU regions): "https://smartapi.vesync.com"
# If device is out of reach, the cloud api sends a timeout response after 7 seconds,
# using 8 here so there is time enough to catch that message
API_BASE_URL_US = 'https://smartapi.vesync.com'
API_BASE_URL_EU = 'https://smartapi.vesync.eu'
REGION_API_MAP = {
    'US': API_BASE_URL_US,
    'EU': API_BASE_URL_EU,
}
NON_EU_COUNTRY_CODES = ['US', 'CA', 'MX', 'JP']
API_TIMEOUT = 8
USER_AGENT = (
    'VeSync/3.2.39 (com.etekcity.vesyncPlatform; build:5; iOS 15.5.0) Alamofire/5.2.1'
)
DEFAULT_TZ = 'America/New_York'
DEFAULT_REGION = 'US'
APP_VERSION = '5.6.60'
APP_ID = 'eldodkfj'
PHONE_BRAND = 'pyvesync'
PHONE_OS = 'Android'
MOBILE_ID = str(randint(1000000000000000, 9999999999999999))  # noqa: S311
USER_TYPE = '1'
BYPASS_APP_V = f'VeSync {APP_VERSION}'
BYPASS_HEADER_UA = 'okhttp/3.12.1'
TERMINAL_ID = '2' + (
    uuid.uuid5(uuid.NAMESPACE_DNS, f'{uuid.getnode():x}-{platform.node() or ""}').hex
)

CLIENT_TYPE = 'vesyncApp'

STATUS_OK = 200


# Generic Constants

KELVIN_MIN = 2700
KELVIN_MAX = 6500


class ProductLines(StrEnum):
    """High level product line."""

    WIFI_LIGHT = 'wifi-light'
    WIFI_AIR = 'wifi-air'
    WIFI_KITCHEN = 'wifi-kitchen'
    SWITCHES = 'Switches'
    WIFI_SWITCH = 'wifi-switch'
    THERMOSTAT = 'thermostat'


class ProductTypes(StrEnum):
    """General device types enum."""

    OUTLET = 'outlet'
    BULB = 'bulb'
    SWITCH = 'switch'
    PURIFIER = 'purifier'
    FAN = 'fan'
    HUMIDIFIER = 'humidifier'
    AIR_FRYER = 'air fryer'
    KITCHEN_THERMOMETER = 'kitchen thermometer'
    THERMOSTAT = 'thermostat'


class IntFlag(IntEnum):
    """DEPRECATED. Integer flag to indicate if a device is not supported.

    This is used by data models as a default value for feature attributes
    that are not supported by all devices.

    The default value is -999.

    Attributes:
        NOT_SUPPORTED: Device is not supported, -999
    """

    NOT_SUPPORTED = -999

    def __str__(self) -> str:
        """Return string representation of IntFlag."""
        return str(self.name)


class StrFlag(StrEnum):
    """DEPRECATED. String flag to indicate if a device is not supported.

    This is used by data models as a default value for feature attributes
    that are not supported by all devices.

    The default value is "not_supported".

    Attributes:
        NOT_SUPPORTED: Device is not supported, "not_supported"
    """

    NOT_SUPPORTED = 'not_supported'


class NightlightStatus(StrEnum):
    """Nightlight status for VeSync devices.

    Values can be converted to int and bool.

    Attributes:
        ON: Nightlight is on.
        OFF: Nightlight is off.
        AUTO: Nightlight is in auto mode.
        UNKNOWN: Nightlight status is unknown.

    Usage:
        ```python
        NightlightStatus.ON
        # <NightlightStatus.ON: 'on'>
        int(NightlightStatus.ON)
        # 1
        bool(NightlightStatus.ON)
        # True
        ```
    """

    ON = 'on'
    OFF = 'off'
    AUTO = 'auto'
    UNKNOWN = 'unknown'

    def __bool__(self) -> bool:
        """Return True if nightlight is on or auto."""
        return self in [NightlightStatus.ON, NightlightStatus.AUTO]

    def __int__(self) -> int:
        """Return integer representation of the enum."""
        if self not in [NightlightStatus.ON, NightlightStatus.OFF]:
            raise ValueError('Only ON and OFF are valid values for int conversion')
        return int(self == NightlightStatus.ON)


class DeviceStatus(StrEnum):
    """VeSync device status enum.

    In addition to converting to int and bool values,
    this enum can be used to convert from bool and int values
    to corresponding string values.

    Attributes:
        ON: Device is on.
        OFF: Device is off.
        PAUSED: Device is paused.
        STANDBY: Device is in standby mode.
        IDLE: Device is idle.
        RUNNING: Device is running.
        UNKNOWN: Device status is unknown.

    Usage:
        ```python
        DeviceStatus.ON
        # <DeviceStatus.ON: 'on'>
        bool(DeviceStatus.ON)
        # True
        int(DeviceStatus.ON)
        # 1
        DeviceStatus.from_int(1)
        # 'on'
        DeviceStatus.from_bool(True)
        # 'on'
        ```
    """

    ON = 'on'
    OFF = 'off'
    PAUSED = 'paused'
    STANDBY = 'standby'
    IDLE = 'idle'
    RUNNING = 'running'
    UNKNOWN = 'unknown'

    def __bool__(self) -> bool:
        """Return True if device is on or running."""
        return self in [DeviceStatus.ON, DeviceStatus.RUNNING]

    def __int__(self) -> int:
        """Return integer representation of the enum.

        1 is ON and RUNNING, 0 is OFF, PAUSED, STANDBY, IDLE.
        -1 is UNKNOWN.
        """
        match self:
            case DeviceStatus.ON | DeviceStatus.RUNNING:
                return 1
            case (
                DeviceStatus.OFF
                | DeviceStatus.PAUSED
                | DeviceStatus.STANDBY
                | DeviceStatus.IDLE
            ):
                return 0
        return -1

    @classmethod
    def from_int(cls, value: int | None) -> str:
        """Convert integer value to corresponding string.

        If value is 1, return ON and if 0, return OFF.
        If value is -999, return NOT_SUPPORTED.
        """
        if value == 1:
            return cls.ON
        if value == 0:
            return cls.OFF
        if value == IntFlag.NOT_SUPPORTED:
            return StrFlag.NOT_SUPPORTED
        return cls.UNKNOWN

    @classmethod
    def from_bool(cls, value: bool | str) -> DeviceStatus:
        """Convert boolean value to corresponding string."""
        if isinstance(value, str):
            return cls.from_bool_string(value)
        return cls.ON if value is True else cls.OFF

    @classmethod
    def from_bool_string(cls, value: str) -> DeviceStatus:
        """Convert boolean value to corresponding string."""
        return cls.ON if value == 'true' else cls.OFF


class ConnectionStatus(StrEnum):
    """VeSync device connection status enum.

    Corresponding boolean value is True if device is online.

    Attributes:
        ONLINE: Device is online.
        OFFLINE: Device is offline.
        UNKNOWN: Device connection status is unknown.

    Methods:
        from_bool(value: bool | None) -> ConnectionStatus:
            Convert boolean value to corresponding string.

    Usage:
        ```python
        ConnectionStatus.ONLINE
        # <ConnectionStatus.ONLINE: 'online'>
        bool(ConnectionStatus.ONLINE)
        # True
        ConnectionStatus.ONLINE == ConnectionStatus.ONLINE
        # True
        ```
    """

    ONLINE = 'online'
    OFFLINE = 'offline'
    UNKNOWN = 'unknown'

    def __bool__(self) -> bool:
        """Return True if device is online."""
        return self == ConnectionStatus.ONLINE

    @classmethod
    def from_bool(cls, value: bool | None) -> ConnectionStatus:
        """Convert boolean value to corresponding string.

        Returns ConnectionStatus.ONLINE if True, else ConnectionStatus.OFFLINE.
        """
        return cls.ONLINE if value else cls.OFFLINE


class NightlightModes(StrEnum):
    """Nightlight modes.

    Attributes:
        ON: Nightlight is on.
        OFF: Nightlight is off.
        DIM: Nightlight is dimmed.
        AUTO: Nightlight is in auto mode.
        UNKNOWN: Nightlight status is unknown.
    """

    ON = 'on'
    OFF = 'off'
    DIM = 'dim'
    AUTO = 'auto'
    UNKNOWN = 'unknown'

    def __bool__(self) -> bool:
        """Return True if nightlight is on or auto.

        Off and unknown are False, all other True.
        """
        return self in [NightlightModes.ON, NightlightModes.AUTO, NightlightModes.DIM]


class ColorMode(StrEnum):
    """VeSync bulb color modes.

    Attributes:
        RGB: RGB color mode.
        HSV: HSV color mode.
        WHITE: White color mode.
        COLOR: Color mode.
    """

    RGB = 'rgb'
    HSV = 'hsv'
    WHITE = 'white'
    COLOR = 'color'


# Purifier Constants


class AirQualityLevel(Enum):
    """Representation of air quality levels as string and integers.

    Attributes:
        EXCELLENT: Air quality is excellent.
        GOOD: Air quality is good.
        MODERATE: Air quality is moderate.
        POOR: Air quality is poor.
        UNKNOWN: Air quality is unknown.

    Methods:
        from_string(value: str | None) -> AirQualityLevel:
            Convert string value to corresponding integer.
        from_int(value: int | None) -> AirQualityLevel:
            Convert integer value to corresponding string.

    Note:
        Alias for "very good" is "excellent".
        Alias for "bad" is "poor".

    Usage:
        ```python
        AirQualityLevels.EXCELLENT
        # <AirQualityLevels.EXCELLENT: 1>
        AirQualityLevels.from_string("excellent")
        # 1
        AirQualityLevels.from_int(1)
        # "excellent"
        int(AirQualityLevels.EXCELLENT)
        # 1
        str(AirQualityLevels.EXCELLENT)
        # "excellent"
        from_string("good")
        # <AirQualityLevels.GOOD: 2>
        from_int(2)
        # "good"
        ```
    """

    EXCELLENT = 1
    GOOD = 2
    MODERATE = 3
    POOR = 4
    UNKNOWN = -1

    def __int__(self) -> int:
        """Return integer representation of the enum."""
        return self.value

    def __str__(self) -> str:
        """Return string representation of the enum."""
        return self.name.lower()

    @classmethod
    def from_string(cls, value: str | None) -> AirQualityLevel:
        """Convert string value to corresponding integer.

        Get enum from string value to normalize different values of the same
        format.

        Note:
            Values are excellent, good, moderate and poor. Aliases are:
            very good for excellent and bad for poor. Unknown is returned
            if value is None or not in the list.
        """
        _string_to_enum = MappingProxyType(
            {
                'excellent': cls.EXCELLENT,
                'very good': cls.EXCELLENT,  # Alias
                'good': cls.GOOD,
                'moderate': cls.MODERATE,
                'poor': cls.POOR,
                'bad': cls.POOR,  # Alias
                'unknown': cls.UNKNOWN,
            }
        )
        if isinstance(value, str) and value.lower() in _string_to_enum:
            return AirQualityLevel(_string_to_enum[value.lower()])
        return cls.UNKNOWN

    @classmethod
    def from_int(cls, value: int | None) -> AirQualityLevel:
        """Convert integer value to corresponding string."""
        if value in [itm.value for itm in cls]:
            return cls(value)
        return cls.UNKNOWN


class PurifierAutoPreference(StrEnum):
    """Preference Levels for Purifier Auto Mode.

    Attributes:
        DEFAULT: Default preference level.
        EFFICIENT: Efficient preference level.
        QUIET: Quiet preference level.
        UNKNOWN: Unknown preference level.
    """

    DEFAULT = 'default'
    EFFICIENT = 'efficient'
    QUIET = 'quiet'
    UNKNOWN = 'unknown'


# Fan Constants


class FanSleepPreference(StrEnum):
    """Sleep mode preferences for VeSync fans.

    Attributes:
        DEFAULT: Default sleep mode.
        ADVANCED: Advanced sleep mode.
        TURBO: Turbo sleep mode.
        EFFICIENT: Efficient sleep mode.
        QUIET: Quiet sleep mode.
        UNKNOWN: Unknown sleep mode.
    """

    DEFAULT = 'default'
    ADVANCED = 'advanced'
    TURBO = 'turbo'
    EFFICIENT = 'efficient'
    QUIET = 'quiet'
    TEMP_SENSE = 'tempSense'
    KIDS = 'kids'
    UNKNOWN = 'unknown'


# Device Features


class Features(StrEnum):
    """Base Class for Features Enum to appease typing."""


# Device Features


class HumidifierFeatures(Features):
    """VeSync humidifier features.

    Attributes:
        ONOFF: Device on/off status.
        CHILD_LOCK: Child lock status.
        NIGHTLIGHT: Nightlight status.
        WATER_LEVEL: Water level status.
        WARM_MIST: Warm mist status.
        AUTO_STOP: Auto stop when target humidity is reached.
            Different from auto, which adjusts fan level to maintain humidity.
    """

    ONOFF = 'onoff'
    CHILD_LOCK = 'child_lock'
    NIGHTLIGHT = 'night_light'
    WATER_LEVEL = 'water_level'
    WARM_MIST = 'warm_mist'
    AUTO_STOP = 'auto_stop'
    NIGHTLIGHT_BRIGHTNESS = 'nightlight_brightness'
    DRYING_MODE = 'drying_mode'


class PurifierFeatures(Features):
    """VeSync air purifier features.

    Attributes:
        CHILD_LOCK: Child lock status.
        NIGHTLIGHT: Nightlight status.
        AIR_QUALITY: Air quality status.
        VENT_ANGLE: Vent angle status.
        LIGHT_DETECT: Light detection status.
        PM25: PM2.5 level status.
        PM10: PM10 level status.
        PM1: PM1 level status.
        AQPERCENT: Air quality percentage status.
        RESET_FILTER: Reset filter status.
    """

    RESET_FILTER = 'reset_filter'
    CHILD_LOCK = 'child_lock'
    NIGHTLIGHT = 'night_light'
    AIR_QUALITY = 'air_quality'
    VENT_ANGLE = 'fan_rotate'
    LIGHT_DETECT = 'light_detect'
    PM25 = 'pm25'
    PM10 = 'pm10'
    PM1 = 'pm1'
    AQPERCENT = 'aq_percent'


class PurifierStringLevels(Features):
    """String levels for Air Purifier fan speed.

    Attributes:
        LOW: Low fan speed.
        MEDIUM: Medium fan speed.
        HIGH: High fan speed.
    """

    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'


class BulbFeatures(Features):
    """VeSync bulb features.

    Attributes:
        ONOFF: Device on/off status.
        DIMMABLE: Dimmable status.
        COLOR_TEMP: Color temperature status.
        MULTICOLOR: Multicolor status.
    """

    ONOFF = 'onoff'
    DIMMABLE = 'dimmable'
    COLOR_TEMP = 'color_temp'
    MULTICOLOR = 'multicolor'


class OutletFeatures(Features):
    """VeSync outlet features.

    Attributes:
        ONOFF: Device on/off status.
        ENERGY_MONITOR: Energy monitor status.
        NIGHTLIGHT: Nightlight status.
    """

    ONOFF = 'onoff'
    ENERGY_MONITOR = 'energy_monitor'
    NIGHTLIGHT = 'nightlight'


class SwitchFeatures(Features):
    """VeSync switch features.

    Attributes:
        ONOFF: Device on/off status.
        DIMMABLE: Dimmable status.
        INDICATOR_LIGHT: Indicator light status.
        BACKLIGHT: Backlight status.
        BACKLIGHT_RGB: RGB backlight status.
    """

    ONOFF = 'onoff'
    DIMMABLE = 'dimmable'
    INDICATOR_LIGHT = 'indicator_light'
    BACKLIGHT = 'backlight'
    BACKLIGHT_RGB = 'backlight_rgb'


class FanFeatures(Features):
    """VeSync fan features."""

    OSCILLATION = 'oscillation'
    SET_OSCILLATION_RANGE = 'set_oscillation_range'
    SOUND = 'sound'
    DISPLAYING_TYPE = 'displaying_type'  # Unknown functionality
    HORIZONTAL_OSCILLATION = 'horizontal_oscillation'
    VERTICAL_OSCILLATION = 'vertical_oscillation'


# Modes


class PurifierModes(Features):
    """VeSync air purifier modes.

    Attributes:
        AUTO: Auto mode.
        MANUAL: Manual mode.
        SLEEP: Sleep mode.
        TURBO: Turbo mode.
        PET: Pet mode.
        UNKNOWN: Unknown mode.
    """

    AUTO = 'auto'
    MANUAL = 'manual'
    SLEEP = 'sleep'
    TURBO = 'turbo'
    PET = 'pet'
    UNKNOWN = 'unknown'


class HumidifierModes(Features):
    """VeSync humidifier modes.

    Attributes:
        AUTO: Auto mode.
        MANUAL: Manual mode.
        HUMIDITY: Humidity mode.
        SLEEP: Sleep mode.
        TURBO: Turbo mode.
        PET: Pet mode.
        UNKNOWN: Unknown mode.
        AUTOPRO: AutoPro mode.
    """

    AUTO = 'auto'
    MANUAL = 'manual'
    HUMIDITY = 'humidity'
    SLEEP = 'sleep'
    TURBO = 'turbo'
    PET = 'pet'
    UNKNOWN = 'unknown'


class FanModes(StrEnum):
    """VeSync fan modes.

    Attributes:
        AUTO: Auto mode.
        NORMAL: Normal mode.
        MANUAL: Manual mode.
        SLEEP: Sleep mode.
        TURBO: Turbo mode.
        PET: Pet mode.
        UNKNOWN: Unknown mode.
        ADVANCED_SLEEP: Advanced sleep mode.
    """

    AUTO = 'auto'
    NORMAL = 'normal'
    MANUAL = 'normal'
    SLEEP = 'sleep'
    TURBO = 'turbo'
    PET = 'pet'
    ECO = 'eco'
    UNKNOWN = 'unknown'


# Air Fryer Constants

AIRFRYER_PID_MAP = {
    'WiFi_SKA_AirFryer137_US': 'wnxwqs76gknqyzjn',
    'WiFi_SKA_AirFryer158_US': '2cl8hmafsthl65bd',
    'WiFi_AirFryer_CS158-AF_EU': '8t8op7pcvzlsbosm',
}
"""PID's for VeSync Air Fryers based on ConfigModule."""


# Thermostat Constants


class ThermostatWorkModes(IntEnum):
    """Working modes for VeSync Aura thermostats.

    Based on the VeSync app and API values.

    Attributes:
        OFF: Thermostat is off (0).
        HEAT: Thermostat is in heating mode (1).
        COOL: Thermostat is in cooling mode (2).
        AUTO: Thermostat is in auto mode (3).
        EM_HEAT: Thermostat is in emergency heating mode (4).
        SMART_AUTO: Thermostat is in smart auto mode (5).
    """

    OFF = 0
    HEAT = 1
    COOL = 2
    AUTO = 3
    EM_HEAT = 4
    SMART_AUTO = 5


class ThermostatFanModes(IntEnum):
    """Fan modes for VeSync Aura thermostats.

    Based on the VeSync app and API values.

    Attributes:
        AUTO: Fan is in auto mode (1).
        ON: Fan is on (2).
        CIRCULATE: Fan is in circulate mode (3).
    """

    AUTO = 1
    ON = 2
    CIRCULATE = 3


class ThermostatHoldOptions(IntEnumMixin):
    """Hold options for VeSync Aura thermostats.

    Attributes:
        UNTIL_NEXT_SCHEDULED_ITEM: Hold until next scheduled item (2).
        TWO_HOURS: Hold for two hours (3).
        FOUR_HOURS: Hold for four hours (4).
        PERMANENTLY: Hold permanently (5).
    """

    UNTIL_NEXT_SCHEDULED_ITEM = 2
    TWO_HOURS = 3
    FOUR_HOURS = 4
    PERMANENTLY = 5


class ThermostatHoldStatus(IntEnumMixin):
    """Set the hold status of the thermostat.

    Attributes:
        SET: Set the hold status (1).
        CANCEL: Cancel the hold status (0).
    """

    SET = 1
    CANCEL = 0


class ThermostatScheduleOrHoldOptions(IntEnumMixin):
    """Schedule or hold options for VeSync Aura thermostats."""

    NOT_SCHEDULE_NOT_HOLD = 0
    ON_SCHEDULE = 1
    ON_HOLD = 2
    VACATION = 3


class ThermostatEcoTypes(IntEnumMixin):
    """Eco types for VeSync Aura thermostats."""

    COMFORT_SECOND = 1
    COMFORT_FIRST = 2
    BALANCE = 3
    ECO_FIRST = 4
    ECO_SECOND = 5


class ThermostatRoutineTypes(IntEnumMixin):
    """Routine types for VeSync Aura thermostats."""

    HOME = 2
    AWAY = 1
    SLEEP = 3
    CUSTOM = 4


class ThermostatAlarmCodes(IntEnumMixin):
    """Alarm codes for VeSync Aura thermostats."""

    HEAT_COOL = 100
    ABOVE_SAFE_TEMP = 102
    HEAT_RUNNING_TIME_TOO_LONG = 104


class ThermostatReminderCodes(IntEnumMixin):
    """Reminder codes for VeSync Aura thermostats."""

    FILTER = 150
    HVAC = 151


class ThermostatWorkStatusCodes(IntEnumMixin):
    """Work status codes for VeSync Aura thermostats."""

    OFF = 0
    HEATING = 1
    COOLING = 2
    EM_HEATING = 3


class ThermostatFanStatus(IntEnumMixin):
    """Fan Status Enum for Aura Thermostats."""

    OFF = 0
    ON = 1


class ThermostatConst:
    """Constants for VeSync Aura thermostats."""

    ReminderCode = ThermostatReminderCodes
    AlarmCode = ThermostatAlarmCodes
    WorkMode = ThermostatWorkModes
    FanStatus = ThermostatFanStatus
    FanMode = ThermostatFanModes
    HoldOption = ThermostatHoldOptions
    EcoType = ThermostatEcoTypes
    RoutineType = ThermostatRoutineTypes
    ScheduleOrHoldOption = ThermostatScheduleOrHoldOptions
    WorkStatus = ThermostatWorkStatusCodes


#  ------------------- AIR FRYER CONST ------------------ #


CUSTOM_RECIPE_ID = 1
CUSTOM_RECIPE_TYPE = 3
CUSTOM_RECIPE_NAME = 'Manual Cook'
CUSTOM_COOK_MODE = 'custom'


# ------------------- OUTLET CONST ------------------ #


class EnergyIntervals(StrEnum):
    """Energy history periods for VeSync outlets.

    Attributes:
        DAILY: Daily energy history.
        WEEKLY: Weekly energy history.
        MONTHLY: Monthly energy history.
        YEARLY: Yearly energy history.
    """

    WEEK = 'week'
    MONTH = 'month'
    YEAR = 'year'


ENERGY_HISTORY_MAP = {
    EnergyIntervals.WEEK: 'getLastWeekEnergy',
    EnergyIntervals.MONTH: 'getLastMonthEnergy',
    EnergyIntervals.YEAR: 'getLastYearEnergy',
}

ENERGY_HISTORY_OFFSET_WHOGPLUG = {
    EnergyIntervals.WEEK: 500000,
    EnergyIntervals.MONTH: 2500000,
}

# ------------------- HUMIDIFIER CONST ------------------ #


class DryingModes(StrEnum):
    """Drying modes for VeSync humidifiers.

    Attributes:
        ON: Drying mode is on.
        OFF: Drying mode is off.
    """

    DONE = 'done'
    RUNNING = 'on'
    PAUSE = 'pause'

    @classmethod
    def from_int(cls, value: int) -> DryingModes | None:
        """Convert integer value to corresponding string.

        If value is 1, return ON and if 0, return OFF.
        If value is -999, return NOT_SUPPORTED.
        """
        if value == 0:
            return cls.DONE
        if value == 1:
            return cls.RUNNING
        if value == 2:  # noqa: PLR2004
            return cls.PAUSE
        return None

    def __int__(self) -> int:
        """Return integer representation of the enum."""
        match self:
            case DryingModes.DONE:
                return 0
            case DryingModes.RUNNING:
                return 1
            case DryingModes.PAUSE:
                return 2


DRYING_MODES: dict[str, int] = {
    DryingModes.DONE: 0,
    DryingModes.RUNNING: 1,
    DryingModes.PAUSE: 2,
}
